# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
The parce Python module.

The main module provides the listed classes and functions, enough to build
a basic language definition or to use the bundled language definitions.

The standard actions that are used by the bundled language definitions to
specify the type of parsed text fragments are in the :mod:`~parce.action`
module. The helper functions for dynamic rule items are in the
:mod:`~parce.rule` module.

It is recommended to import *parce* like this::

    import parce

although in a language definition it can be easier to do this::

    from parce import Language, lexicon, skip, default_action, default_target
    from parce.rule import words, bygroup   # whichever you need
    import parce.action as a

Then you get the ``Language`` class and ``lexicon`` decorator from parce, and
all standard actions can be accessed via the ``a`` prefix, like ``a.Text``.

.. py:data:: version

   The version as a three-tuple(major, minor, patch). See :mod:`~parce.pkginfo`.

.. py:data:: version_string

   The version as a string.

"""

# imported when using from parce import *
__all__ = (
    # important classes
    'Cursor',
    'Document',

    # often used names when defining languages
    'default_action',
    'default_target',
    'Language',
    'lexicon',
    'skip',

    # toplevel functions
    'events',
    'find',
    'root',
    'theme_by_name',
    'theme_from_file',
    'tokens',
)

from . import document, lexer, rule, ruleitem, treebuilder, treedocument, util
from .lexicon import lexicon
from .language import Language
from .document import Cursor
from .pkginfo import version, version_string


class Document(treedocument.TreeDocumentMixin, document.Document):
    """A Document that automatically keeps its contents tokenized.

    You can specify your own TreeBuilder. By default, a BackgroundTreeBuilder
    is used.

    """
    def __init__(self, root_lexicon=None, text="", builder=None):
        document.Document.__init__(self, text)
        if builder is None:
            builder = treebuilder.BackgroundTreeBuilder(root_lexicon)
        else:
            builder.root.clear()
            builder.root.lexicon = root_lexicon
        treedocument.TreeDocumentMixin.__init__(self, builder)
        if text:
            builder.rebuild(text)


def find(name=None, *, filename=None, mimetype=None, contents=None):
    """Find a root lexicon, either by language name, or by filename, mimetype
    and/or contents.

    If you specify a name, tries to find the language with that name, ignoring
    the other arguments.

    If you don't specify a name, but instead one or more of the other (keyword)
    arguments, tries to find the language based on filename, mimetype or
    contents.

    If a language is found, returns the root lexicon. If no language could be
    found, None is returned (which can also be used as root lexicon, resulting
    in an empty token tree).

    Examples::

        >>> import parce
        >>> parce.find("xml")
        Xml.root
        >>> parce.find(contents='{"key": 123;}')
        Json.root
        >>> parce.find(filename="style.css")
        Css.root

    This function uses the :mod:`~parce.registry` module and by default it
    finds all bundled languages. See the module's documentation to find out how
    to add your own languages to a registry.

    """
    from . import registry
    if name:
        lexicon_name = registry.find(name)
    else:
        for lexicon_name in registry.suggest(filename, mimetype, contents):
            break
        else:
            return
    if lexicon_name:
        return registry.root_lexicon(lexicon_name)


def root(root_lexicon, text):
    """Return the root context of the tree structure of all tokens from text."""
    return treebuilder.build_tree(root_lexicon, text)


def tokens(root_lexicon, text):
    """Convenience function that yields all the tokens from the text."""
    return root(root_lexicon, text).tokens()


def events(root_lexicon, text):
    """Convenience function that yields all the events from the text."""
    return lexer.Lexer([root_lexicon]).events(text)


def theme_by_name(name="default"):
    """Return a Theme from the default themes in the themes/ directory."""
    from .theme import Theme
    return Theme.byname(name)


def theme_from_file(filename):
    """Return a Theme loaded from the specified CSS filename."""
    from .theme import Theme
    return Theme(filename)


# these can be used in rules where a pattern is expected
default_action = util.Symbol("default_action")   #: denotes a default action for unmatched text
default_target = util.Symbol("default_target")   #: denotes a default target when no text matches


skip = ruleitem.SkipAction()
"""A dynamic action that yields no tokens, thereby ignoring the matched text."""





# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
This module defines classes and functions to build a tree structure by parsing
a text string.

To get the tree of tokens using a particular root lexicon from a string
of text, use :func:`build_tree`.

A more advanced approach is using the :class:`TreeBuilder`, which can build a
tree in one go as well, but is also capable of updating an existing tree when
the text changes on a particular position, e.g. while typing in a text editor.
In this case, tokens in front of the modified region are reused (carefully
checking whether changes affect earlier regions), and also tokens at the end of
the modified region are reused, if they have the same context ancestry.

TreeBuilder also reports the start and end position of the updated region, and
the lexicons that were left open at the end, which in some languages can mean
that a document or a certain structure is incomplete.

The TreeBuilder is designed so that it is possible to perform tokenizing in a
background thread, and even interrupt tokenizing when changes are to be applied
while processing previous changes.

The :class:`BackgroundTreeBuilder` provides an implementation using Python
threads.

"""

import operator
import threading

from parce.lexer import Lexer
from parce.util import Observable, tokens
from parce.target import TargetFactory
from parce.tree import Context, make_tokens
from parce.treebuilderutil import (
    BuildResult, ReplaceResult, Changes, get_prepared_lexer, new_tree)


def build_tree(root_lexicon, text, pos=0):
    """Build and return a tree in one go."""
    from parce.tree import Context, make_tokens # local is faster
    root = context = Context(root_lexicon, None)
    lexer = Lexer([root_lexicon])
    for e in lexer.events(text, pos):
        if e.target:
            for _ in range(e.target.pop, 0):
                context = context.parent
            for lexicon in e.target.push:
                context = Context(lexicon, context)
                context.parent.append(context)
        context.extend(make_tokens(e, context))
    return root


class TreeBuilder(Observable):
    """Build a tree from parsing the text.

    The root node of the tree is in the ``root`` instance attribute.
    This root context is never replaced, although its lexicon may change and
    of course its children.

    Call :meth:`rebuild` to build or rebuild the tree. This method stores the
    desired changes to the tree and calls :meth:`start_processing`, which can
    be re-implemented to support asynchronous tree building.

    The actual building of a tree happens in :meth:`build_new_tree` which
    builds a (replacement) tree without making any changes yet to the current
    tree.

    The result of :meth:`build_new_tree` is a tuple of arguments that is used
    used when calling :meth:`replace_tree`, which integrates the updated
    subtree in the main tree structure. This method sets three instance
    attributes:

    ``start``, ``end``:
        indicate the region the tokens were changed. After build(), start is
        always 0 and end = len(text), but after rebuild(), these values
        indicate the range that was actually re-tokenized.

    ``lexicons``:
        the list of open lexicons (excluding the root lexicon) at the end of
        the document. This way you can see in which lexicon parsing ended.

        If a tree was rebuilt, and old tail tokens were reused, the lexicons
        variable is not set, meaning that the old value is still valid. If the
        TreeBuilder was not used before, lexicons is an empty tuple.

    No other variables or state are kept, so if you don't need the above
    information anymore, you can throw away the TreeBuilder after use.

    During the building process, the TreeBuilder emits certain events you can
    subscribe to, using the :meth:`~parce.util.Observable.connect` method
    provided by the :class:`~parce.util.Observable` class that's mixed into
    this TreeBuilder class.

    The following events are emitted, with following arguments:

    ``"started"``:
        emitted when a (re)build starts; the handler is called without
        arguments

    ``"replace"``:
        emitted just before the tree actually changes (while the new tree is
        being built, the tree is still unchanged and accessible, but between
        the ``"replace"`` and ``"finished"`` events the tree is in an
        inconsistent state)

    ``"finished"``:
        emitted when a (re)build has finished; the handler is called without
        arguments

    ``"updated"``:
        emitted when a (re)build has finished; the handler is called with two
        arguments: ``start``, ``end``, that denote the updated range

    ``"peek"``:
        emitted by the default implementation of the :meth:`peek` method,
        the handler is called with two arguments: ``start``, ``tree``

    ``"invalidate"``:
        emitted by the default implementation of the :meth:`invalidate_context`
        method, the handler is called with the Context that needs to be
        invalidated

    For example, to get notified when a build process starts::

        >>> b = TreeBuilder(MyLang.root)
        >>> def hi_there():
        ...     print("started")
        ...
        >>> b.connect("started", hi_there)
        >>> b.rebuild("some boring text")
        started
        >>>

    """
    start = 0
    end = 0
    lexicons = ()

    peek_threshold = 0  #: set to a value > 0 to get :meth:`peek` called during building

    def __init__(self, root_lexicon=None):
        super().__init__()
        self.root = Context(root_lexicon, None)
        self.busy = False
        self.changes = []

    def tree(self, text):
        """Convenience method to build a tree and return the root node."""
        self.rebuild(text)
        return self.get_root(True)

    def rebuild(self, text, root_lexicon=False, start=0, removed=0, added=None):
        """Tokenize the modified part of the text again and update the tree.

        The arguments:

        ``text``
            The text to parse. Always give the entire text, also when you only
            actually changed a small part. The tree builder needs to check text
            before and after the changed region, and possibly re-parse more
            text.

        ``root_lexicon``
            The root lexicon to use (default: False). False means no change;
            can be None or any Lexicon. If not False, the tree is always
            rebuilt completely.

        ``start``
            Position of the change (default: 0)

        ``removed``
            The number of removed characters (default: 0)

        ``added``
            The number of added characters (default: None, which means
            all characters from start to the end of the text)

        Calls :meth:`build_new_tree` and :meth:`replace_tree` to do the actual work.

        """
        if added is None:
            added = len(text) - start
        self.lock(True)
        self.changes.append((text, root_lexicon, start, removed, added))
        self.lock(False)
        if not self.busy:
            self.busy = True
            self.start_processing()

    def build_new_tree(self, text, root_lexicon, start, removed, added):
        """Build a new tree without yet modifying the current tree.

        Tokens from the current tree are reused as much as possible. From
        tokens at the tail (after the end of the modified region) the pos
        attribute is updated if necessary.

        The arguments:

        ``text``
            The text to parse. Always the entire text, also when only a small
            portion was changed.

        ``root_lexicon``
            The root lexicon to use. False means no change; can be None or any
            Lexicon. If not False, the tree is always rebuilt completely.

        ``start``
            Position of the change.

        ``removed``
            The number of removed characters.

        ``added``
            The number of added characters.

        Returns a ``Result`` five-tuple with ``tree``, ``start``, ``end``,
        ``offset`` and ``lexicons`` values. The ``start`` and ``end`` are the
        insert positions in the old tree.

        The new ``tree`` is intended to replace a part of, or the whole old
        tree. If ``start`` == 0 and ``lexicons`` is not None; the whole tree
        can be replaced. (In this case; the root lexicon might have changed!)
        Use :meth:`replace_tree` to insert the result tree in the old tree.

        If ``start`` > 0, tokens in the old tree before start are to be
        preserved.

        If ``lexicons`` is None, old tail tokens after ``end`` must be reused,
        and the old list of open lexicons is still relevant. The ``offset`` then
        gives the position change for the tokens that are reused.

        """
        from parce.tree import Context, make_tokens # local is faster

        if root_lexicon is not False:
            start, removed, added = 0, 0, len(text)
        else:
            root_lexicon = self.root.lexicon

        # manage end, and record if there is text after the modified part (tail)
        end = start + removed

        # record the position change for tail tokens that may be reused
        offset = added - removed

        if not root_lexicon:
            return BuildResult(Context(root_lexicon, None), start, start + added, 0, [])

        # If there remains text after the modified part,
        # we try to reuse the old tokens
        tail = False
        if start + added < len(text):
            # find the first token after the modified part
            tail_token = self.root.find_token_after(end)
            if tail_token:
                tail_gen = ((t, t.pos) for t in tail_token.forward_including()
                        if not t.group)
                for tail_token, tail_pos in tail_gen:
                    tail = True
                    break

        lowest_start = start
        changes = self.changes
        tree = None
        while True:
            # when restarting, see if we can reuse (part of) the new tree
            if tree:
                result = get_prepared_lexer(tree, text, start)
                if result:
                    lexer, events, tokens = result
                    t = tokens[0]
                    context = t.parent
                    for p, i in t.ancestors_with_index():
                        del p[i+1:]
                    del context[-1]
                else:
                    tree = None
            # find insertion spot in old tree
            if not tree:
                start = min(lowest_start, start)
                result = get_prepared_lexer(self.root, text, start)
                if result:
                    lexer, events, tokens = result
                    t = tokens[0]
                    context, tree = new_tree(t)
                    start = tokens[-1].end
                    lowest_start = min(lowest_start, start)
                else:
                    tree = context = Context(root_lexicon, None)
                    lexer = Lexer([root_lexicon])
                    events = lexer.events(text)
                    lowest_start = 0
                peek = self.peek_threshold + lowest_start if self.peek_threshold else 0
            # start parsing
            for e in events:
                if e.target:
                    for _ in range(e.target.pop, 0):
                        context = context.parent
                    for lexicon in e.target.push:
                        context = Context(lexicon, context)
                        context.parent.append(context)
                tokens = make_tokens(e, context)
                if tail:
                    # handle tail
                    pos = tokens[0].pos - offset
                    if pos > tail_pos:
                        for tail_token, tail_pos in tail_gen:
                            if tail_pos >= pos:
                                break
                        else:
                            tail = False
                    if pos == tail_pos and tokens[0].equals(tail_token):
                        # we can reuse the tail from tail_pos
                        return BuildResult(tree, lowest_start, tail_pos, offset, None)
                context.extend(tokens)
                if changes:
                    # handle changes
                    c = self.get_changes()
                    if c:
                        # break out and adjust the current tokenizing process
                        text = c.text
                        start = c.start
                        if c.root_lexicon != False:
                            root_lexicon = c.root_lexicon
                            if not root_lexicon:
                                return BuildResult(Context(root_lexicon, None), 0, len(text), 0, [])
                            start = 0
                            tail = False
                            tree = None

                        elif tail:
                            # reuse old tail?
                            new_tail_pos = start + c.added
                            if new_tail_pos >= len(text):
                                tail = False
                            else:
                                offset += c.added - c.removed
                                new_tail_pos -= offset
                                if new_tail_pos > tail_pos:
                                    for tail_token, tail_pos in tail_gen:
                                        if tail_pos >= new_tail_pos:
                                            break
                                    else:
                                        tail = False
                        break # break the for loop to restart at new start pos
                if peek:
                    if tokens[-1].pos > peek:
                        # call peek with a copy of the current tree.
                        copied_tree = tree.copy()
                        # remove the first empty context
                        t = copied_tree[0]
                        while t and t.is_context:
                            t = t[0]
                        while t.is_context and not t:
                            t = t.parent
                            del t[0]
                        self.peek(lowest_start, copied_tree)
                        peek = 0
            else:
                # we ran till the end, also return the open lexicons
                return BuildResult(tree, lowest_start, len(text), 0, lexer.lexicons[1:])
        raise RuntimeError("shouldn't come here")

    def replace_tree(self, result):
        """Modify the tree using the result from :meth:`build_new_tree`.

        In most types of GUI applications, this method should be called in the
        main (GUI) thread.

        The changes are delegated to the various ``replace_`` methods, which
        can be reimplemented to get fine-grained monitoring of and control over
        the tree-replacing process.

        Additionally, this method calls :meth:`invalidate_context` with the
        youghest Context that had children removed or added.

        """
        tree, start, end, offset, lexicons = result

        if not tree.lexicon or tree.lexicon != self.root.lexicon:
            # whole tree update
            root = self.root
            for n in tree:
                n.parent = root
            self.replace_nodes(self.root, slice(None), tree)
            self.replace_root_lexicon(tree.lexicon)
            self.invalidate_context(self.root)

        else:

            context = self.root
            start_trail = self.root.find_token_left_with_trail(start)[1] if start else []
            end_trail = self.root.find_token_with_trail(end)[1] if lexicons is None else []

            # find the context that really changes, adjust trails
            i = 0
            for i, (s, e) in enumerate(zip(start_trail, end_trail)):
                if s == e and tree.is_context and len(tree) == 1:
                    context = context[s]
                    tree = tree[0]
                else:
                    break
            del start_trail[:i], end_trail[:i]

            for n in tree:
                n.parent = context

            slice_end = None
            if end_trail:
                # join stuff after end_trail with tree
                c = context
                t = tree
                end_trail[-1] -= 1
                for i in end_trail:
                    l = len(t) - 1
                    if slice_end is None:
                        slice_end = i + 1
                    else:
                        s = c[i+1:]
                        t.extend(s)
                        for n in s:
                            n.parent = t
                        if offset:
                            for n in tokens(s):
                                n.pos += offset
                    if t:
                        t = t[l]    # t can be empty if len(end_trail) == 1, is last iteration anyway
                    c = c[i]

            if offset and slice_end is not None and slice_end < len(context):
                self.replace_pos(context, slice(slice_end, None), offset)

            if start_trail:
                # replace stuff after start_trail with tree
                c = context
                t = tree
                for i in start_trail[:-1]:
                    for n in t[1:]:
                        n.parent = c
                    self.replace_nodes(c, slice(i + 1, slice_end), t[1:])
                    slice_end = None
                    t = t[0]
                    c = c[i]
                i = start_trail[-1]
                for n in t:
                    n.parent = c
                self.replace_nodes(c, slice(i + 1, slice_end), t)
                self.invalidate_context(c)
            else:
                self.replace_nodes(context, slice(slice_end), tree)
                self.invalidate_context(context)

            if offset:
                for p, i in context.ancestors_with_index():
                    self.replace_pos(p, slice(i + 1, None), offset)

        return ReplaceResult(start, end + offset, lexicons)

    def replace_nodes(self, context, slice_, nodes):
        """Replace the context's slice with new nodes.

        This method is called by :meth:`replace_tree`.
        You can reimplement this method to notify others of the change.

        """
        context[slice_] = nodes

    def replace_root_lexicon(self, lexicon):
        """Set the root lexicon.

        This method is called by :meth:`replace_tree`.
        You can reimplement this method to notify others of the change.

        """
        self.root.lexicon = lexicon

    def replace_pos(self, context, slice_, offset):
        """Adjust the pos attribute of all tokens in the context's slice.

        This method is called by :meth:`replace_tree`.
        You can reimplement this method to notify others of the change.

        """
        for t in tokens(context[slice_]):
            t.pos += offset

    def invalidate_context(self, context):
        """Called with the younghest Context that had children are removed or
        added.

        This means that the meaning of this context probably has changed, for
        example when you want to transform the context to some other data
        structure, and that the ancestors also need to be invalidated.

        The default implementation of this method emits the ``invalidate``
        event, see :meth:`~parce.util.Observable.connect`.

        """
        self.emit("invalidate", context)

    def get_root(self, wait=False, callback=None, args=(), kwargs={}):
        """Return the root element of the completed tree.

        This is simply the ``root`` instance attribute, but this method only
        returns the tree when the ``busy`` attribute is False.

        If wait is True, this call blocks until tokenizing is done, and the
        full tree is returned. If wait is False, None is returned if the tree
        is still busy being built.

        If a callback is given and tokenizing is still busy, that callback is
        called once when tokenizing is ready. If given, args and kwargs are the
        arguments the callback is called with, defaulting to () and {},
        respectively.

        Note that, for the lifetime of a TreeBuilder, the root element is always
        the same. The root element is also accessible in the `root` attribute.
        But using this method you can be sure that you are dealing with a
        complete and fully intact tree.

        """
        if self.busy:
            if callback:
                if args or kwargs:
                    callback = lambda: callback(*args, **kwargs)
                self.connect("finished", callback, True)
            if not wait:
                return
            self.wait()
        return self.root

    def get_changes(self):
        """Get and combine the stored change requests in a Changes object.

        This may only be called from the same thread that also performs the
        :meth:`rebuild()`.

        """
        c = Changes()
        while self.changes:
            c.add(*self.changes.pop(0))
        return c

    def start_processing(self):
        """Called when there are recorded changes to process.

        The default implementation read all build stages from the
        :meth:`process` generator until exhausted. You can inherit from this
        method to call it e.g. in a background thread.

        """
        for stage in self.process():
            pass

    def process(self):
        """Process all changes and update the tree.

        This method behaves as a generator coroutine, instead of simply calling
        this method, you should iterate over its output, which reports which
        stage the process is at.

        Yields "build" when about to build a new tree; "replace" when about to
        replace a new tree; (which can be repeated); "finish" when finished
        looping, and "done" at the very end.

        When re-implementing :meth:`start_processing`, you can choose to decide
        which stages are to be run in a background thread and which in a main
        (GUI) thread.

        You should exhaust the generator fully.

        """
        self.process_started()
        start = end = -1
        lexicons = False    # no change
        self.lock(True)
        c = self.get_changes()
        while c and c.has_changes():
            self.lock(False)
            yield "build"
            result = self.build_new_tree(c.text, c.root_lexicon, c.start, c.removed, c.added)
            yield "replace"
            self.emit("replace")
            r = self.replace_tree(result)
            start = r.start if start == -1 else min (start, r.start)
            end = r.end if end == -1 else max(c.new_position(end), r.end)
            if r.lexicons is not None:
                lexicons = r.lexicons
            self.lock(True)
            c = self.get_changes()
        yield "finish"
        if start != -1:
            self.start = start
        if end != -1:
            self.end = end
        if lexicons is not False:
            self.lexicons = lexicons
        self.busy = False
        self.lock(False)
        self.process_finished()
        yield "done"

    def wait(self):
        """Implement to wait for completion if a background job is running.

        The default implementation does nothing, and immediately returns.

        """
        pass

    def peek(self, start, tree):
        """This is called from :meth:`build_new_tree` with a sneak preview tree.

        This can be used to get a small tree before the new tree is built
        completely, which is useful to update e.g. highlighting of a small
        portion of a document that is edited by a user, instead of waiting on
        the whole tree to update (which may cause slow highlighting updates).

        When build_new_tree (the build stage) is called from a background
        thread, this method will also be called from that same thread.

        Enable the ``peek()`` feature by setting the ``peek_threshold``
        attribute to a value > 0. E.g. the value 1000 will cause the
        :meth:`peek` method to be called with a tree that encompasses at least
        1000 characters (starting with the start position).

        The tree that is given, is a copy of the current tree. It is safe to
        use it in another thread, although its contents are not valid anymore
        when the build has finished, or when a build is restarted, causing
        peek() to be called a second time. (A build is restarted when there are
        new changes close to the position the build originally started.)

        The default implementation of this method emits the ``peek`` event, see
        :meth:`~parce.util.Observable.connect`.

        """
        self.emit("peek", start, tree)

    def lock(self, acquire):
        """Acquire lock (True) or release lock (False). Does nothing by default.

        If you want to run the full update and replace jobs in a background
        thread, you may need locking, to prevent changes from going unnoticed.

        """
        pass

    def process_started(self):
        """Called at the start ot the tree building process.

        The default implementation of this method emits the ``started`` event,
        see :meth:`~parce.util.Observable.connect`.

        """
        self.emit("started")

    def process_finished(self):
        """Called when tree building is done.

        The default implementation of this method emits the ``updated(start,
        end)`` and ``finished`` events, see
        :meth:`~parce.util.Observable.connect`.

        """
        self.emit("updated", self.start, self.end)
        self.emit("finished")


class BackgroundTreeBuilder(TreeBuilder):
    """A TreeBuilder that can tokenize a text in a Python thread.

    In BackgroundTreeBuilder, :meth:`rebuild()` returns immediately, because
    :meth:`start_processing()` has been reimplemented to call itself in a
    background thread.

    You can continue adding changes while previous changes are processed;
    the tree builder will immediately adapt to the new changes.

    To be sure you get a completed tree, call ``get_root(True)``.

    """
    def __init__(self, root_lexicon=None):
        super().__init__(root_lexicon)
        self.job = None
        self._lock = threading.Lock()

    def lock(self, acquire):
        """Reimplemented to actually lock/unlock."""
        self._lock.acquire() if acquire else self._lock.release()

    def start_processing(self):
        """Reimplemented to call start_processing in a background thread."""
        self.job = threading.Thread(target=super().start_processing)
        self.job.start()

    def wait(self):
        """Reimplemented to await our background thread if active."""
        job = self.job
        if job:
            job.join()

    def process_finished(self):
        """Reimplemented to clear the job attribute."""
        self.job = None
        super().process_finished()





# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Registry of language definitions.

Instead of importing language definitions directly, you can use a Registry to
manage and find language definitions.

The registry stores the fully qualified name for a root lexicon, for example
``"parce.lang.css.Css.root"``. This qualified name should have at least 2 dots,
to separate module name, class name and the name of the root lexicon.

Use :func:`find` to find a language definition by name, or :func:`suggest` to
find a language definition for a particular file type. There is basic
functionality to pick a language definition based on file name, mime type
and/or the contents of the file.

Using the :func:`register` function it is possible to register your own
language definitions at runtime and make them available through parce.
As a service, the bundled languages in ``parce.lang`` are automatically
registered in the global registry.

The global registry is in the :attr:`registry` module variable.
You can also create and populate your own :class:`Registry`.

"""


import collections
import fnmatch
import importlib
import itertools
import operator
import re


Item = collections.namedtuple("Item", (
    "name",
    "desc",
    "aliases",
    "filenames",
    "mimetypes",
    "guesses",
))


class Registry(dict):
    """Registry of language definitions.

    The ``Registry`` is based on the Python dictionary class, and maps fully
    qualified lexicon names (such as ``"parce.lang.css.Css.root"``) to
    :class:`Item` tuples.

    """
    def register(self, lexicon_name, *,
        name,
        desc,
        aliases = [],
        filenames = [],
        mimetypes = [],
        guesses = [],
    ):
        """Register or update a Language's root lexicon for a particular filename
        (patterns), particular mime types or based on contents of the file.

        The arguments:

        ``lexicon_name``
            The fully qualified name of a root lexicon, e.g.
            ``"parce.lang.css.Css.root"``. Must contain at least two dots.
        ``name``
            A human-readable name for the file type
        ``desc``
            A short description
        ``aliases``
            An optional list of other names this lexicon can be found under.
        ``filenames``
            A list of tuples (pattern, weight). A pattern is a plain filename or a
            filename with globbing characters, e.g. ``"Makefile"`` or ``"*.c"``,
            and the weight is a floating point value indicating the probability
            that the root lexicon should be chosen for this filename (0..1 range).
        ``mimetypes``
            A list of tuples (mimetype, weight). A mimetype is a string like
            ``"text/css"``, the weight is a floating point value indicating the
            probability that the root lexicon should be chosen for this filename
            (0..1 range).
        ``guesses``
            A list of tuples (regexp, weight). The first 5000 characters of the
            contents are matched against the regular expression, and when it
            matches, the weight is added to the already computed weight for this
            root lexicon.

        This method simply creates an :class:`Item` tuple with all the arguments
        and stores it using the lexicon name as key.

        """
        item = Item(name, desc, aliases, filenames, mimetypes, guesses)
        self[lexicon_name] = item

    def suggest(self, filename=None, mimetype=None, contents=None):
        """Return a list of registered language definitions, sorted on relevance.

        The filename has the most weight, if two have the same weight, the mimetype
        is looked at; if still the same, the contents are looked at with some
        heuristic.

        Every item in the returned list is the fully qualified name of the root
        lexicon, e.g. ``"parce.lang.css.Css.root"``.

        """
        weights = collections.defaultdict(int)
        if filename:
            for name in self:
                weight = max((w for pat, w in self[name].filenames
                               if fnmatch.fnmatch(filename, pat)), default=0)
                if weight:
                    weights[name] += weight
        if mimetype:
            for name in self:
                weight = max((w for mtype, w in self[name].mimetypes
                               if mtype == mimetype), default=0)
                if weight:
                    weights[name] += weight

        # check the contents only if no filename/mimetype matched
        # or there were multiple matches with the same weight
        if weights:
            names = sorted(weights, key=weights.get, reverse=True)
            if len(names) == 1 or weights[names[0]] > weights[names[1]]:
                return names
        else:
            names = self.keys()
        if contents:
            contents = contents[:5000]
            for name in names:
                weight = sum(w for regex, w in self[name].guesses
                               if re.search(regex, contents))
                if weight:
                    weights[name] += weight
        return sorted(weights, key=weights.get, reverse=True)

    def find(self, name):
        """Find a fully qualified lexicon name for the specified name.

        First, tries to find the exact match on the ``name`` attribute, then
        the aliases, then a case insensitive match, and then the same for the
        Language class name.

        """
        aliases = []
        nocases = []
        classes = []
        name_lowered = name.lower()
        for lexicon_name, item in self.items():
            if name == item.name:
                return lexicon_name
            if name in item.aliases:
                aliases.append(lexicon_name)
            if name_lowered == item.name.lower() or any(
                    alias.lower() == name_lowered for alias in item.aliases):
                nocases.append(lexicon_name)
            cls = lexicon_name.rsplit(".", 2)[1]
            if name_lowered == cls.lower():
                classes.append(lexicon_name)
        for lexicon_name in itertools.chain(aliases, nocases, classes):
            return lexicon_name


# the global Registry is in the ``registry`` module variable
registry = Registry()


def register(lexicon_name, **kwargs):
    """:meth:`~Registry.register` a lexicon in the global registry."""
    registry.register(lexicon_name, **kwargs)


def suggest(filename=None, mimetype=None, contents=None):
    """:meth:`~Registry.suggest` zero or more lexicons from the global registry."""
    return registry.suggest(filename, mimetype, contents)


def find(name):
    """:meth:`~Registry.find` a lexicon by name from the global registry."""
    return registry.find(name)


def root_lexicon(lexicon_name):
    """Import the module and return the root lexicon.

    Eg, for the ``lexicon_name`` ``"parce.lang.css.Css.root"`` imports the
    ``parce.lang.css`` module and returns the ``Css.root`` lexicon.

    """
    module, cls, root = lexicon_name.rsplit(".", 2)
    mod = importlib.import_module(module)
    return getattr(getattr(mod, cls), root)




## register the bundled languages

import parce.lang._registry



# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
Parse GNU Texinfo.

https://www.gnu.org/software/texinfo/

"""

__all__ = ('Texinfo',)

import re

from parce import Language, lexicon, default_action
from parce.action import (
    Bracket, Comment, Delimiter, Escape, Keyword, Name, Verbatim, Text)
from parce.rule import bygroup, ifgroup


class Texinfo(Language):

    @lexicon
    def root(cls):
        yield r'@[@{}. ]', Escape
        yield r'''@['",=^`~](\{[a-zA-Z]\}|[a-zA-Z]\b)''', Escape
        yield r'@c(?:omment)?\b', Comment, cls.singleline_comment
        yield r'@ignore\b', Comment, cls.multiline_comment
        yield r'@verbatim\b', Keyword.Verbatim, cls.verbatim
        yield r'@html', Keyword, cls.html
        yield r'(@lilypond)\b(?:(\[)([^\n\]]*)(\]))?(\{)?', bygroup(
            Name.Function, Bracket, Name.Property, Bracket, Bracket.Start), \
            ifgroup(5, cls.lilypond_brace, cls.lilypond_block)
        yield r'(@[a-zA-Z]+)(?:(\{)(\})?)?', bygroup(
                ifgroup(2, ifgroup(3, Name.Symbol, Name.Function), Name.Command),
                Bracket.Start,
                Bracket.End), \
            ifgroup(2, ifgroup(3, (), cls.brace))
        yield default_action, Text

    @lexicon
    def brace(cls):
        yield r'\}', Bracket.End, -1
        yield from cls.root

    @lexicon
    def verbatim(cls):
        yield r'(@end)[ \t]+(verbatim)\b', bygroup(Keyword, Keyword.Verbatim), -1
        yield default_action, Verbatim

    @lexicon
    def html(cls):
        from .html import Html
        yield r'(@end)[ \t]+(html)\b', bygroup(Keyword, Keyword), -1
        yield from Html.root

    @lexicon
    def lilypond_block(cls):
        from .lilypond import LilyPond
        yield r'(@end)[ \t]+(lilypond)\b', bygroup(Keyword, Name.Function), -1
        yield from LilyPond.root

    @lexicon
    def lilypond_brace(cls):
        from .lilypond import LilyPond
        yield r'\}', Bracket.End, -1
        yield from LilyPond.root

    #---------- comments ------------------------
    @lexicon(re_flags=re.MULTILINE)
    def singleline_comment(cls):
        yield '$', None, -1
        yield from cls.comment_common()

    @lexicon
    def multiline_comment(cls):
        yield r'@end\s+ignore\b', Comment, -1
        yield from cls.comment_common()




# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
This module defines a fairly large amount of standard actions.
Most are used by the bundled language definitions and may also be used
or inherited from by your own language definitions.

The themes built into *parce* also use the names of these standard actions
to determine the visual appearance of parsed text pieces.

See for a full listing :doc:`stdactions`. See the :mod:`~parce.standardaction`
module for more information about standard actions.

"""

from . import standardaction

# Note: the marked part below of this file is literally shown in the
# documentation (stdactions.rst)

# BEGIN_ACTIONS

a = standardaction.StandardAction

# Base actions, derive from these:

Text            = a("Text")
Whitespace      = a("Whitespace")
Keyword         = a("Keyword")
Delimiter       = a("Delimiter")
Name            = a("Name")             # see below for many derived actions
Literal         = a("Literal")          # see below for many derived actions
Comment         = a("Comment")


# Mixin actions, yield subtle style changes:

Alert           = a("Alert")
Important       = a("Important")
Unimportant     = a("Unimportant")
Special         = a("Special")

Definition      = a("Definition")       # a thing is defined (vs referred to)
Invalid         = a("Invalid")          # invalid input
Escape          = a("Escape")           # escaped text like \n in a string
Template        = a("Template")         # a template for something else
Pseudo          = a("Pseudo")           # e.g. a pseudo-class or -comment
Preprocessed    = a("Preprocessed")

Inserted        = a("Inserted")         # inserted text (e.g. in a diff)
Deleted         = a("Deleted")          # deleted text (e.g. in a diff)
Quoted          = a("Quoted")           # e.g. a blockquote
Bold            = a("Bold")             # Bold text
Emphasized      = a("Emphasized")       # Emphasized text


# Mixin actions that are not styled by default:

Start           = a("Start")            # start of something
End             = a("End")              # end of something
Indent          = a("Indent")           # denotes an indent
Dedent          = a("Dedent")           # denotes a dedent


# Actions that derive of Name:

Name.Attribute
Name.Builtin
Name.Class
Name.Command
Name.Constant
Name.Decorator
Name.Entity
Name.Exception
Name.Function
Name.Identifier
Name.Macro
Name.Markup
Name.Method
Name.Namespace
Name.Object
Name.Property
Name.Symbol
Name.Tag
Name.Type
Name.Variable


# Actions that derive of Literal:

Data            = Literal.Data
Verbatim        = Literal.Verbatim
String          = Literal.String        # a quoted string
Character       = Literal.Character     # a single character
Number          = Literal.Number        # a numeric value
Fraction        = Number.Fraction       # a fraction/rational value
Boolean         = Number.Boolean        # a boolean value
Literal.Color
Literal.Email
Literal.Url
Literal.Timestamp
Literal.Input
Literal.Output
Literal.Error
Literal.Prompt


# Actions that derive of String:

String.Double
String.Single


# Actions that derive of Delimiter:

Connection      = Delimiter.Connection  # connects things
Direction       = Delimiter.Direction   # implicates a direction
Bracket         = Delimiter.Bracket     # for { }, ( ), [ ] etc.
ModeChange      = Delimiter.ModeChange  # a special char that changes mode
Operator        = Delimiter.Operator    # an arithmetic operator
Quote           = Delimiter.Quote       # a quote, e.g. " or «
Separator       = Delimiter.Separator   # a separator, like \\ or | in LaTeX

Operator.Assignment                     # e.g. = to denote an assignment


# END_ACTIONS
del a, standardaction



# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
The Formatter uses a Theme to highlight text according to the token's
``action`` attribute. The action is mapped to a TextFormat by the theme.

When a MetaTheme is used, a Formatter is capable of switching Theme
as the MetaTheme provides a special theme for a certain embedded language.

All kinds of text formatting and/or highlighting can be implemented by using
or inheriting of Formatter. If you need to convert the TextFormats from the
theme to something else, you can provide a factory to Formatter to do that.

If you need more special behaviour, you can inherit from Formatter and
reimplement ``format_ranges()``, to do other things or to use a different
FormatContext.


"""


import sys
import collections
import contextlib
import threading
import weakref


from . import util


FormatRange = collections.namedtuple("FormatRange", "pos end textformat")


class FormatCache:
    """A FormatCache caches conversions from TextFormat to something else.

    It can keep a window() TextFormat into account for sub-themes (themes
    that are invoked on a per-language basis), and it can adjust the formats
    for the sub-themed tokens with the window() format for that theme.

    The factory that was given to the Formatter is used by the FormatCache to
    convert the TextFormat of the theme to something used by the formatter.

    A Formatter keeps a FormatCache for its theme, and in the case of a
    MetaTheme, a FormatCache is used for every sub-theme. The FormatContext
    is used to switch theme, and in its default implementation it switches
    the FormatCache it uses.

    """
    def __init__(self, theme, factory, add_window=True):
        """Caches conversions from TextFormat to something else."""
        if add_window:
            window = theme.baseformat()
            self.window = factory(window)
            self._format = lambda action: factory(window + theme.textformat(action))
        else:
            self.window = None
            self._format = lambda action: factory(theme.textformat(action))

    @util.cached_method
    def format(self, action):
        """Return our text format for the action, caching it for later reuse."""
        return self._format(action)


class FormatContext:
    """A FormatContext is used during a formatting job.

    It maintains the state needed to format using a MetaTheme. The only API
    used is the ``push()`` and ``pop()`` method, which are called by the
    ``tokens()`` method of MetaTheme.

    The default Formatter uses the ``window`` property and ``format`` method
    to get the window style for the current theme and the factory that turns
    a standard action if the format we want.

    """
    def __init__(self, formatter):
        self._stack = []
        self._formatter = formatter
        self._current_theme = None
        self._set_theme(formatter.theme())

    def _switch_language(self, language):
        self._set_theme(self._formatter.theme().get_theme(language))

    def _set_theme(self, theme):
        if theme is not self._current_theme:
            self._current_theme = theme
            c = self.cache = self._formatter.format_cache(theme)
            self.window = c.window
            self.format = c.format

    def push(self, language):
        self._stack.append(language)
        if len(self._stack) < 2 or language is not self._stack[-2]:
            self._switch_language(language)

    def pop(self):
        language = self._stack.pop()
        if len(self._stack) > 0 and language is not self._stack[-1]:
            self._switch_language(self._stack[-1])


class Formatter:
    """A Formatter is used to format or highlight text according to a theme.

    Supply the theme, and an optional factory that converts a TextFormat
    to something else.

    """
    def __init__(self, theme, factory=None):
        self._lock = threading.Lock()   # lock for FormatCaches
        self._theme = theme
        self._factory = factory or (lambda f: f)
        self._caches = weakref.WeakKeyDictionary()
        self._caches[theme] = c = FormatCache(theme, factory, False)

    def theme(self):
        """Return the Theme we were instantiated with."""
        return self._theme

    @util.cached_method
    def baseformat(self, role="window", state="default"):
        """Return our textformat for the current line."""
        return self._factory(self._theme.baseformat(role, state))

    def format_cache(self, theme):
        """Return a FormatCache for the Theme.

        The FormatCache caches the converted textformat, optionally taking
        the default window style into account.  And the Formatter caches the
        FormatCaches :-)

        """
        try:
            return self._caches[theme]
        except KeyError:
            with self._lock:
                try:
                    return self._caches[theme]
                except KeyError:
                    add_window = self.theme().get_add_window(theme)
                    c = self._caches[theme] = FormatCache(
                                    theme, self._factory, add_window)
                return c

    def format_ranges(self, tree, start=0, end=None):
        """Yield FormatRange(pos, end, format) three-tuples.

        The ``format`` is the value returned by Theme.textformat() for the
        token's action, converted by our factory (and cached of course).
        Ranges with a TextFormat for which our factory returns None are
        skipped.

        """
        c = FormatContext(self)
        def stream():
            prev_end = sys.maxsize
            for t in self._theme.tokens(c, tree, start, end):
                if t.pos > prev_end and c.window is not None:
                    # if a sub-language is active, draw its background
                    yield prev_end, t.pos, c.window
                f = c.format(t.action)
                if f is not None:
                    yield t.pos, t.end, f
                prev_end = t.end
        yield from util.merge_adjacent(stream(), FormatRange)





# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""A CSS parser.

| CSS3 Syntax:     https://www.w3.org/TR/css-syntax-3/
| Selector syntax: https://www.w3.org/TR/selectors-4/

We also use this parser inside parce, to be able to store default
highlighting formats in css files.

"""

__all__ = ('Css', 'CssTransform', 'Rule', 'Atrule', 'Color', 'Value')

import collections
import re

from parce import Language, lexicon, skip, default_action, default_target
from parce.action import (
    Bracket, Comment, Delimiter, Escape, Invalid, Keyword, Literal, Name,
    Number, Operator, String)
from parce.rule import TEXT, bygroup, ifmember, ifeq
from parce.transform import Transform


RE_CSS_ESCAPE = r"\\(?:[0-9A-Fa-f]{1,6} ?|.)"
RE_CSS_NUMBER = (
    r"[+-]?"               # sign
    r"(?:\d*\.)?\d+"       # mantisse
    r"(?:[Ee][+-]\d+)?")   # exponent
RE_CSS_IDENTIFIER_LA = r"(?=-?(?:[^\W\d]|\\[0-9A-Fa-f])|--)" # lookahead
RE_CSS_IDENTIFIER = (
    r"(?:-?(?:[^\W\d]+|" + RE_CSS_ESCAPE + r")|--)"
    r"(?:[\w-]+|" + RE_CSS_ESCAPE + r")*")
RE_CSS_AT_KEYWORD = r"@" + RE_CSS_IDENTIFIER
# match either 8, 6, 4 or 3 hex digits
RE_HEX_COLOR = r"#[0-9a-fA-F]{3}(?:[0-9a-fA-F]{5}|[0-9a-fA-F]{3}|[0-9a-fA-F]?)"


class Css(Language):
    @lexicon
    def root(cls):
        yield from cls.toplevel()

    @classmethod
    def toplevel(cls):
        yield r"@", Keyword, cls.atrule, cls.atrule_keyword
        yield r"/\*", Comment, cls.comment
        yield r"\s+", skip  # skip whitespace
        yield default_target, cls.prelude

    @lexicon
    def prelude(cls):
        yield r"\{", Bracket, -1, cls.rule
        yield r"(?=</)", None, -1   # back off if HTML </style> tag follows...
        yield from cls.selectors()

    @classmethod
    def selectors(cls):
        """Yield selectors, used in prelude and selector_list."""
        yield r"\s+", skip              # skip whitespace
        yield r"[>+~]|\|\|", Operator   # combinators
        yield r",", Delimiter           # comma
        yield r"/\*", Comment, cls.comment
        yield r'"', String, cls.dqstring
        yield r"'", String, cls.sqstring
        yield default_target, cls.selector

    @lexicon
    def selector(cls):
        yield r"\*", Keyword    # "any" element
        yield r"\|", Keyword    # css selector namespace prefix separator
        yield r"#", Name.Identifier.Definition, cls.id_selector
        yield r"\.(?!\d)", Keyword, cls.class_selector
        yield r"::", Keyword, cls.pseudo_element
        yield r":", Keyword, cls.pseudo_class
        yield r"\[", Delimiter, cls.attribute_selector, cls.attribute
        yield RE_CSS_IDENTIFIER_LA, None, cls.element_selector
        yield default_target, -1

    @lexicon
    def selector_list(cls):
        """The list of selectors in :is(bla, bla), etc."""
        yield r"\)", Delimiter, -2  # also leave the pseudo_class context
        yield from cls.selectors()

    @lexicon
    def rule(cls):
        """Declarations of a qualified rule between { and }."""
        yield r"\}", Bracket, -1
        yield from cls.inline()

    @lexicon
    def inline(cls):
        """CSS that would be in a rule block, but also in a HTML style attribute."""
        yield RE_CSS_IDENTIFIER_LA, None, cls.declaration, cls.property
        yield from cls.common()

    @lexicon
    def declaration(cls):
        """A property: value;  declaration."""
        yield r":", Delimiter
        yield r";", Delimiter, -1
        yield r"!important\b", Keyword, -1
        yield from cls.common()
        yield r"\s+", skip  # stay here on whitespace only
        yield default_target, -1

    @classmethod
    def common(cls):
        """Stuff that can be everywhere."""
        yield r'"', String, cls.dqstring
        yield r"'", String, cls.sqstring
        yield r"/\*", Comment, cls.comment
        yield r"\{", Bracket, cls.rule
        yield RE_CSS_NUMBER, Number, cls.unit
        yield RE_HEX_COLOR, Literal.Color
        yield r"(url)(\()", bygroup(Name, Delimiter), cls.url_function
        # an ident-token is found using a lookahead pattern, the whole ident-
        # token is in the identifier context
        yield RE_CSS_IDENTIFIER_LA, None, cls.identifier
        yield r"[:,;@%!]", Delimiter

    @lexicon
    def unit(cls):
        """Unit directly after a number, e.g. the ``px`` in 100px, also ``%``."""
        yield "%", Operator.Percent, -1
        yield from cls.identifier_common(Name.Unit)

    # ------------ selectors for identifiers in different roles --------------
    @classmethod
    def identifier_common(cls, action):
        """Yield an ident-token and give it the specified action."""
        yield RE_CSS_ESCAPE, Escape
        yield r"[\w-]+", action
        yield default_target, -1

    @lexicon
    def element_selector(cls):
        """A tag name used as selector."""
        yield from cls.identifier_common(Name.Tag)

    @lexicon
    def property(cls):
        """A CSS property."""
        from .css_words import CSS3_ALL_PROPERTIES
        yield from cls.identifier_common(
            ifmember(TEXT, CSS3_ALL_PROPERTIES, Name.Property.Definition, Name.Property))

    @lexicon
    def attribute(cls):
        """An attribute name."""
        yield from cls.identifier_common(Name.Attribute)

    @lexicon
    def id_selector(cls):
        """#id"""
        yield from cls.identifier_common(Name.Identifier.Definition)

    @lexicon
    def class_selector(cls):
        """.classname"""
        yield from cls.identifier_common(Name.Class)

    @lexicon
    def attribute_selector(cls):
        """Stuff between [ and ]."""
        yield r"\]", Delimiter, -1
        yield r"[~|^$*]?=", Operator
        yield r'"', String, cls.dqstring
        yield r"'", String, cls.sqstring
        yield RE_CSS_IDENTIFIER_LA, None, cls.ident_token
        yield r'\s+', skip
        yield default_action, Invalid

    @lexicon
    def pseudo_class(cls):
        """Things like :first-child etc."""
        yield r"\(", Delimiter, cls.selector_list
        yield from cls.identifier_common(Name.Class.Pseudo)

    @lexicon
    def pseudo_element(cls):
        """Things like ::first-letter etc."""
        yield from cls.identifier_common(Name.Tag.Pseudo)

    # --------------------- @-rule ------------------------
    @lexicon
    def atrule(cls):
        """Contents following '@'."""
        yield r"\{", Bracket, cls.atrule_block
        yield from cls.atrule_common()

    @lexicon
    def atrule_nested(cls):
        """An atrule that has nested toplevel contents (@media, etc.)"""
        yield r"\{", Bracket, cls.atrule_nested_block
        yield from cls.atrule_common()

    @lexicon
    def atrule_keyword(cls):
        """The first identifier word in an @-rule."""
        yield r"(media|supports|document)\b", Keyword, -1, cls.atrule_nested
        yield from cls.identifier_common(Keyword)

    @lexicon
    def atrule_block(cls):
        """a { } block from an @-rule."""
        yield r"\}", Bracket, -2  # immediately leave the atrule context
        yield from cls.inline()

    @lexicon
    def atrule_nested_block(cls):
        """a { } block from @media, @document or @supports."""
        yield r"\}", Bracket, -2  # immediately leave the atrule_nested context
        yield from cls.toplevel()

    @classmethod
    def atrule_common(cls):
        """Common stuff inside @-rules."""
        yield r";", Delimiter, -1
        yield r":", Keyword, cls.pseudo_class
        yield from cls.common()
        yield r'(?=</)', None, -1   # leave atrule when </style tag follows

    @lexicon
    def ident_token(cls):
        """An ident-token where quoted or unquoted text is allowed."""
        yield from cls.identifier_common(Name.Symbol)

    @lexicon
    def identifier(cls):
        """An ident-token that could be a color or a function()."""
        from .css_words import CSS3_NAMED_COLORS
        yield r"\(", Delimiter, cls.function
        yield from cls.identifier_common(ifeq(TEXT, "transparent", Literal.Color,
            ifmember(TEXT, CSS3_NAMED_COLORS, Literal.Color, Name.Symbol)))

    @lexicon
    def function(cls):
        """Contents between identifier( ... )."""
        yield r"\)", Delimiter, -2  # go straight out of the identifier context
        yield r"\(", Delimiter, 1
        yield from cls.common()
        yield r"[*/+-]", Operator

    @lexicon
    def url_function(cls):
        """url(....)"""
        yield r"\)", Delimiter, -1
        yield r'"', String, cls.dqstring
        yield r"'", String, cls.sqstring
        yield r"/\*", Comment, cls.comment
        yield RE_CSS_ESCAPE, Escape
        yield default_action, Literal.Url

    @lexicon
    def dqstring(cls):
        """A double-quoted string."""
        yield r'"', String, -1
        yield from cls.string()

    @lexicon
    def sqstring(cls):
        """A single-quoted string."""
        yield r"'", String, -1
        yield from cls.string()

    @classmethod
    def string(cls):
        """Common rules for string."""
        yield default_action, String
        yield RE_CSS_ESCAPE, String.Escape
        yield r"\\\n", String.Escape
        yield r"\n", Invalid, -1

    @lexicon
    def comment(cls):
        """A comment."""
        yield r"\*/", Comment, -1
        yield from cls.comment_common()


class CssTransform(Transform):
    r"""Transform a CSS stylesheet into a simpler data structure.

    The data structure created by this Transform only contains plain Python
    strings, lists, (named) tuples, dictionaries and :class:`Value` objects.
    :class:`Value` represents any item from a property value list. Most
    notably, a CSS color (hexadecimal, named or via the CSS
    ``rgb()``/``rgba()`` functions) is converted to a named :class:`Color`
    four-tuple.

    A tree created by the Css.root lexicon becomes a list of tuples
    that are either a Rule or an Atrule named tuple.

    An :class:`Atrule` named tuple corresponds to an @-rule and consists of
    three items, the ``keyword`` which contains the name, the ``contents``,
    which contains all items after the name, and the ``block`` which contains
    the part between { and }.

    There are three at-rule types:

    1. Nested at-rule: the ``keyword`` is either ``"media"``, ``"supports"`` or
       ``"document"``, the ``contents`` contains the query strings and Value
       instances; the ``block`` contains the list of nested Rule/Atrule tuples.

    2. At-rules with a properties block, like ``@page:left { ... }``: the
       keyword can be anything, the block is a dictionary of properties like
       the ``properties`` dictionary of a normal Rule (see below). The
       ``contents`` contains the stuff between the block and the initial
       keyword.

    3. At-rules without a block; like ``@import url("filename.css");``, the
       ``block`` is None for these at-rules.

    A :class:`Rule` named tuple corresponds to a normal CSS rule and consists
    of two items, the ``prelude``, which contains the selectors, and the
    ``properties`` dictionary.

    The *prelude* is a list of selector groups. Each selector group is also
    a list, containing at least one selector dictionary and optionally
    operator strings and more selector dictionaries. See the :meth:`prelude`
    method.

    The *properties* is a dictionary mapping property names to lists of
    :class:`Value` instances. A Value can express any CSS property value, like
    a quoted string, unquoted name, number with or without unit, url, etc. It
    also recognizes named, `rgb/rgba` and hexadecimal colors, which can be
    found as a :class:`Color` tuple in the :attr:`Value.color` attribute. It
    does not yet parse ``calc()`` function calls.

    For example::

        >>> from parce import root
        >>> from parce.transform import transform_tree
        >>> from parce.lang.css import Css
        >>> transform_tree(root(Css.root, 'h1 { color: red; }'))
        [Rule(prelude=[[{'element_selector': ['h1']}]], properties={'color':
        [<Value text='red', color=Color(r=255, g=0, b=0, a=1.0)>]})]

    """
    def root(self, items):
        """Return a list of Rule or Atrule tuples."""
        result = []
        prelude = None
        for name, obj in items.items():
            if name == "prelude":
                prelude = obj
            elif name == "rule":
                result.append(Rule(prelude, obj))
                prelude = None
            elif name == "atrule":
                result.append(obj)
        return result

    def prelude(self, items):
        r"""Return a Css prelude.

        A prelude is a list of selector lists. A Css prelude that contains a
        comma has more than one selector lists.

        A selector list is a list of selector dictionaries with possible
        combinator operators in between. The operators can be: ``" "`` (space),
        ``">"``, ``"~"``, ``"+"``, or ``"||"``.

        Every selector is a dictionary, and inbetween are operator strings. A
        comma in the selector causes the prelude to contain more than one list.
        Every selector list consists of selector dicts with an operator or
        whitespace in between.

        """
        # skip the { that starts the rule which is normally there
        if items and items[-1] == '{':
            items = items[:-1]
        prelude = []
        result = []
        for i in items:
            if i.is_token:
                if i == ',':
                    prelude.append(result)
                    result = []
                elif i.action is Operator:
                    if result and isinstance(result[-1], dict):
                        # dont append operator at start, or two operators
                        result.append(i.text)
            elif i.name == "selector":
                if result and isinstance(result[-1], dict):
                    # append descendant combinator if no operator was there
                    result.append(" ")
                result.append(i.obj)
        if result:
            prelude.append(result)
        # cleanup prelude: remove operators that are at the end of a selector
        # list
        for selectors in prelude:
            if selectors and not isinstance(selectors[-1], dict):
                del selectors[-1]
        # remove empty selector lists
        prelude = [selectors for selectors in prelude if selectors]
        return prelude

    def selector(self, items):
        """Return a dictionary object.

        The possible keys are: "element_selector", "id_selector",
        "class_selector", "pseudo_class", "pseudo_element",
        "attribute_selector".

        If present, the value is a list of objects created by that context.
        Most objects are simple strings, but for pseudo_class it is a (name,
        selector_list) tuple, and for attribute_selector it is a four-tuple of
        the contents between the ``[`` and ``]``.

        "*" is ignored, '|' is not yet handled.

        """
        d = collections.defaultdict(list)
        for name, obj in items.items():
            d[name].append(obj)
        return dict(d)

    def selector_list(self, items):
        """Stuff inside :not(), :is(), etc."""
        # skip the closing ) which is normally there
        if items and items[-1] == ')':
            items = items[:-1]
        return self.prelude(items)

    def rule(self, items):
        """A Css rule, between { ... }."""
        return self.inline(items)

    def inline(self, items):
        """Return a dictionary of the property values."""
        d = {}
        for i in items.items("declaration"):
            if i.obj:
                prop, values = i.obj
                d[prop] = values
        return d

    def declaration(self, items):
        """Return a two-tuple(property, value).

        The value is a list of Value instances from :meth:`common`.

        """
        items = iter(items)
        for i in items:
            if not i.is_token and i.name == "property":
                propname = i.obj
                values = list(self.common(i
                        for i in items if i not in (':', ';')))
                return propname, values

    def unit(self, items):
        """Return the name of the unit in itens."""
        if items and items[0] == "%":
            return '%'
        return self.get_ident_token(items)[0]

    def element_selector(self, items):
        """Return the name of the element_selector."""
        return self.get_ident_token(items)[0]

    def property(self, items):
        """Return the name of the property."""
        return self.get_ident_token(items)[0]

    def attribute(self, items):
        """Return the name of the attribute."""
        return self.get_ident_token(items)[0]

    def id_selector(self, items):
        """Return the name of the id_selector."""
        return self.get_ident_token(items)[0]

    def class_selector(self, items):
        """Return the name of the class_selector."""
        return self.get_ident_token(items)[0]

    def attribute_selector(self, items):
        """Return a four-tuple representing the contents between [ and ].

        The tuple: (attribute, operator, value, flag).

        """
        attr = op = val = flag = None
        for i in items:
            if i.is_token:
                if i.action is Operator:
                    op = i.text
            elif i.name in ('sqstring', 'dqstring'):
                val = i.obj
            elif i.name == 'attribute':
                attr = i.obj
            elif i.name == 'ident_token':
                if val:
                    flag = i.obj
                else:
                    val = i.obj
        return  attr, op, val, flag

    def pseudo_class(self, items):
        """Return a tuple(name, selector_list).

        The ``name`` is the name of the pseudo class, the selector_list
        is a list of selectors like the ``prelude`` of a rule. For pseudo
        classes without arguments, the selector_list is None.

        """
        name = self.get_ident_token(items)[0]
        selector_list = None
        if items and not items[-1].is_token and items[-1].name == "selector_list":
            selector_list = items[-1].obj
        return name, selector_list

    def pseudo_element(self, items):
        """Return the name of the pseudo element."""
        return self.get_ident_token(items)[0]

    def atrule(self, items):
        """Return a Atrule named tuple."""
        if items and not items[0].is_token and items[0].name == "atrule_keyword":
            keyword = items.pop(0).obj
        else:
            keyword = None
        block = None
        for n, i in enumerate(items):
            if not i.is_token:
                if i.name == "atrule_nested":
                    contents, block = i.obj
                    break
                elif i.name == "atrule_block":
                    block = i.obj   # the properties dict
                    contents = tuple(self.common(items[:n-1]))  # skip {
                    break
            elif i == ';':
                contents = tuple(self.common(items[:n]))
                break
        else:
            contents = tuple(self.common(items))
        return Atrule(keyword, contents, block)

    def atrule_nested(self, items):
        """Return a two-tuple: the stuff before the nested block and the nested block."""
        nested = None
        if items and not items[-1].is_token and items[-1].name == "atrule_nested_block":
            nested = items.pop().obj
            items.pop() # skip {
        return tuple(self.common(items)), nested

    def atrule_keyword(self, items):
        """Return the name of the atrule keyword."""
        return self.get_ident_token(items)[0]

    def atrule_block(self, items):
        """Return the properties dict in an atrule block."""
        return self.inline(items)

    def atrule_nested_block(self, items):
        """Return a list of Rule or Atrule tuples."""
        return self.root(items)

    def ident_token(self, items):
        """Return the ident_token."""
        return self.get_ident_token(items)[0]

    def identifier(self, items):
        """Return a Value.

        For a color name, returns a Value with a color, otherwise
        a Value with the text.

        If the identifier also has a function sub-context, a Value representing
        a function call is returned (this is done by the
        :func:`get_css_function_call` function, which is capable of
        interpreting ``url``, ``rgb`` and ``rgba`` function calls).

        """
        text, action = self.get_ident_token(items)
        if items and not items[-1].is_token and items[-1].name == "function":
            funcargs = items[-1].obj
            return self.get_css_function_call(text, funcargs)
        from .css_words import CSS3_NAMED_COLORS
        if action is Literal.Color or text in CSS3_NAMED_COLORS:
            color = self.get_named_color(text)
            return Value(color=color, text=text)
        return Value(text=text)

    def function(self, items):
        """Return a list of Value instances and delimiting tokens."""
        # skip the closing ) which is normally there
        if items and items[-1] == ')':
            items = items[:-1]
        return tuple(self.common(items))

    def url_function(self, items):
        """Return a Value with the url."""
        def gen():
            for i in items:
                if i.is_token:
                    if i.action is Escape:
                        yield self.get_escape(i.text)
                    elif i.action is Literal.Url:
                        yield i.text
                elif i.name in ('dqstring', 'sqstring'):
                    yield i.obj
        return Value(url=''.join(gen()))

    def dqstring(self, items):
        """Return the contents of a double-quoted string."""
        if items and items[-1] == '"':
            items = items[:-1]
        return ''.join(self.get_string(items))

    def sqstring(self, items):
        """Return the contents of a single-quoted string."""
        if items and items[-1] == "'":
            items = items[:-1]
        return ''.join(self.get_string(items))

    ### we don't implement comment, so all comments are ignored

    ### helper methods
    def common(self, items):
        """Yield any values, see Css.common().

        Every item is either a Value instance or a delimiting token.
        If an ``identifier`` context is followed by a ``function`` context,
        they are combined into a Value with funcname and arguments by the
        :meth:`get_css_function_call` method.

        """
        items = iter(items)
        i = next(items, None)
        while i is not None:
            if i.is_token:
                if i.action is Number:
                    value = Value(text=i.text, number=self.get_number(i.text))
                    i = next(items, None)
                    if i and not i.is_token and i.name == "unit":
                        value.unit = i.obj
                        i = next(items, None)
                    yield value
                    continue
                elif i.action is Literal.Color:
                    # a hexadecimal color (a named color is an identifier)
                    color = self.get_hex_color(i.text[1:])
                    yield Value(color=color, text=i.text)
                elif i.action is Keyword or i.action in Delimiter:
                    yield i.text
                elif i.action is Name and i.text == "url":
                    next(items, None)   # skip (
            elif i.name in ('sqstring', 'dqstring'):
                yield Value(text=i.obj, quoted=True)
            elif i.name in (
                    'url_function',
                    'identifier',
                    'pseudo_class',
                    'attribute',
                    'selector_list',
                 ):
                yield i.obj
            i = next(items, None)

    def get_css_function_call(self, name, arguments):
        """Return a Value for a CSS function call. Handles rgb/rgba."""
        if name in ('rgb', 'rgba'):
            return Value(color=self.get_rgba_color(arguments))
        elif name == "url":
            # normally handled by url_function
            text = ''.join(v.text for v in arguments
                    if isinstance(v, Value) and v.text is not None)
            return Value(url=text)
        return Value(funcname=name, arguments=arguments)

    def get_number(self, text):
        """Get the value of a Number."""
        num = float(text)
        if num.is_integer():
            num = int(num)
        return num

    def get_escape(self, text):
        """Get the value of escaped text."""
        value = text[1:]
        if value == '\n':
            return ''
        try:
            codepoint = int(value, 16)
        except ValueError:
            return value
        return chr(codepoint)

    def get_string(self, items):
        """Yield the parts of a string.

        Called by :meth:`sqstring` and :meth:`dqstring`.

        """
        for i in items.tokens():
            if i.action is String:
                yield i.text
            elif i.action is String.Escape:
                yield self.get_escape(i.text)

    def get_ident_token(self, items):
        """Return a two-tuple(name, action).

        Combines tokens in an identifier context, (see :meth:`Css.identifier_common`).

        """
        actions = [None]
        def gen():
            for i in items.tokens():
                if i.action is Escape:
                    yield self.get_escape(i.text)
                elif i.action not in Delimiter:
                    if i.action in Name:
                        actions.append(i.action)
                    yield i.text
        name = ''.join(gen())
        return name, actions[-1]

    def get_hex_color(self, text):
        """Return a named four-tuple Color(r, g, b, a) describing color and alpha.

        The ``text`` is a hexadecimal color without the hash, like "FA0042".
        The r, g, b values are in the range 0..255, a in 0..1; 1.0 is fully
        opaque.

        """
        r, g, b, a = -1, -1, -1, 1.0
        l = len(text)
        c = int(text, 16)
        if l == 3:
            # 17F -> 1177FF
            r = (c // 256 & 15) * 17
            g = (c // 16 & 15) * 17
            b = (c & 15) * 17
        elif l == 4:
            # fourth digit is alpha value
            r = (c // 4096 & 15) * 17
            g = (c // 256 & 15) * 17
            b = (c // 16 & 15) * 17
            a = (c & 15) * 17 / 255
        elif l == 6:
            # six digits
            r = c // 65536 & 255
            g = c // 256 & 255
            b = c & 255
        else: # l == 8:
            # eight digits: last two is alpha value
            r = c // 16777216 & 255
            g = c // 65536 & 255
            b = c // 256 & 255
            a = (c & 255) / 255
        return Color(r, g, b, a)

    def get_named_color(self, text):
        """Return a named four-tuple Color(r, g, b, a) describing color and alpha.

        The ``text`` is a CSS3 color name.
        The r, g, b values are in the range 0..255, a in 0..1; 1.0 is fully
        opaque.

        """
        if text == "transparent":
            r, g, b, a = 0, 0, 0, 0
        else:
            r, g, b, a = -1, -1, -1, 1.0
            from .css_words import CSS3_NAMED_COLORS
            try:
                r, g, b = CSS3_NAMED_COLORS[text]
            except KeyError:
                pass
        return Color(r, g, b, a)

    def get_rgba_color(self, func_args):
        """Convert the arguments to a rgba(1 2 3 4) call to a Color."""
        values = (v for v in func_args if isinstance(v, Value) and v.number)
        r = next(values, -1)
        if r != -1: r = self.get_number_value(r, 255)
        g = next(values, -1)
        if g != -1: g = self.get_number_value(g, 255)
        b = next(values, -1)
        if b != -1: b = self.get_number_value(b, 255)
        a = next(values, 1.0)
        if a != 1.0: a = self.get_number_value(a, 1.0)
        return Color(r, g, b, a)

    def get_number_value(self, value, maximum):
        """Return a numeric value from the Value object.

        If the value object has a percentage, apply it so 100% yields the
        maximum value. Otherwise, the number is constrained to be in the
        0..maximum range.

        """
        n = value.number
        if value.unit == "%":
            n = n * maximum // 100
        if n < 0: n = 0
        if n > maximum: n = maximum
        return n


#: An at-rule. For nested atrules the nested stylesheet is in a list ``block``,
#: for other at-rules that end with a rule with properties, the properties
#: dict is in ``block``; when there is no block, ``block`` is None.
Atrule = collections.namedtuple("Atrule", "keyword contents block")

#: A normal rule
Rule = collections.namedtuple("Rule", "prelude properties")

#: A named tuple holding the (r, g, b, a) value of a color.
Color = collections.namedtuple("Color", "r g b a")


class Value:
    """Any value that can occur in a CSS property declaration.

    The value of a CSS property is always a list of Value instances.

    For a *numerial* value, the ``number`` attribute contains the numeric value,
    and the ``text`` attribute the textual representation as present in the CSS
    file. A unit (or "``%``") that was specified, is in the ``unit`` attribute.

    For a *color* value, the color is in the ``color`` attribute as a
    :class:`Color` four-tuple. When a CSS3 named color was specified, the name
    of the color is in the ``text`` attribute, or when a hexadecimal color was
    specified, the hexadecimal notation is also in the ``text`` attribute.

    For a value that can either be a *quoted string* or an *ident_token*, the
    value is in the ``text`` attribute. If it originally was a quoted string,
    the ``quoted`` attribute is set to True.

    If the value represents a *URL* specified via the ``url()`` function, the
    URL is in the ``url`` attribute.

    If the value represents a *function call*, the name of the function is in
    the ``funcname`` attribute, and the argument list in the ``arguments``
    attribute. Except for the ``url()``, the ``rgb()`` and ``rgba()``
    functions, which are handled by the CssTransform class.

    """
    def __init__(self,
            text = None,
            number = None,
            unit = None,
            url = None,
            color = None,
            funcname = None,
            quoted = None,
            arguments = (),
            ):
        self.text = text
        self.number = number
        self.unit = unit
        self.url = url
        self.color = color
        self.funcname = funcname
        self.quoted = quoted
        self.arguments = list(arguments)

    def __repr__(self):
        def gen():
            for name, value in self.__dict__.items():
                if value not in (None, []):
                    yield '{}={}'.format(name, repr(value))
        return '<{} {}>'.format(self.__class__.__name__, ', '.join(gen()))





# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Parse XML.

"""

__all__ = ('Dtd', 'Xml')

import re

from parce import Language, lexicon, skip, default_action
from parce.action import (
    Bracket, Comment, Data, Delimiter, Escape, Invalid, Keyword, Name,
    Operator, String, Text, Whitespace)
from parce.rule import (
    MATCH, TEXT, bygroup, call, dselect, ifgroup, select, words)


# source: https://www.w3.org/TR/xml/#NT-NameStartChar
RE_XML_NAME_START_CHAR = (
    '_:A-Za-z\xC0-\xD6\xD8-\xF6'
    '\xF8-\u02FF\u0370-\u037D\u037F-\u1FFF'
    '\u200C\u200D\u2070-\u218F\u2C00-\u2FEF'
    '\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD'
    '\U00010000-\U000EFFFF'
)
RE_XML_NAME_CHAR = '-.0-9\xB7\u0300-\u036F\u203F-\u2040' + RE_XML_NAME_START_CHAR
RE_XML_NAME = _N_ = fr'[{RE_XML_NAME_START_CHAR}][{RE_XML_NAME_CHAR}]*'
RE_XML_NAME_TOKEN = _T_ = fr'[{RE_XML_NAME_CHAR}]*'

class _XmlBase(Language):
    """Common stuff between Xml and Dtd."""
    @lexicon
    def dqstring(cls):
        yield r'&\S*?;', String.Escape
        yield default_action, String.Double
        yield r'"', String.Double.End, -1

    @lexicon
    def sqstring(cls):
        yield r'&\S*?;', String.Escape
        yield default_action, String.Single
        yield r"'", String.Single.End, -1

    @lexicon
    def comment(cls):
        yield r'-->', Comment.End, -1
        yield r'--', Comment.Invalid
        yield from cls.comment_common()

    @classmethod
    def common_defs(cls):
        """Common stuff inside DOCTYPE or ENTITY declarations etc."""
        yield from cls.find_strings()
        yield fr'%{_N_};', Name.Entity.Escape

    @classmethod
    def find_strings(cls):
        yield r'"', String.Double.Start, cls.dqstring
        yield r"'", String.Single.Start, cls.sqstring

    @classmethod
    def find_comments(cls):
        yield r'<!--', Comment.Start, cls.comment


class Xml(_XmlBase):
    """Parse XML."""
    @lexicon(re_flags=re.IGNORECASE)
    def root(cls):
        yield from cls.find_comments()
        yield r'(<!\[)(CDATA)(\[)', bygroup(Delimiter, Data.Definition, Delimiter), cls.cdata
        yield fr'(<!)(DOCTYPE)\b(?:\s*({_N_}))?', \
            bygroup(Delimiter, Keyword, Name.Tag.Definition), cls.doctype
        yield fr'(<\?)({_N_})?', bygroup(Bracket.Preprocessed.Start, Name.Tag.Preprocessed), \
            cls.processing_instruction
        yield fr'(<\s*?/)\s*({_N_})\s*(>)', bygroup(Delimiter, Name.Tag, Delimiter), -1
        yield fr'(<)\s*({_N_})(?:\s*((?:/\s*)?>))?', \
            bygroup(Delimiter, Name.Tag, Delimiter), dselect(MATCH(3), {
                None: cls.attrs,        # no ">" or "/>": go to attrs
                ">": cls.tag,           # a ">": go to tag
            })                          # by default ("/>"): stay in context
        yield r'&\S*?;', Escape
        yield default_action, select(call(str.isspace, TEXT), Text, Whitespace)

    @lexicon
    def tag(cls):
        yield from cls.root()

    @lexicon
    def cdata(cls):
        yield default_action, Data
        yield r'\]\]>', Delimiter, -1

    @lexicon
    def processing_instruction(cls):
        yield fr'({_N_})\s*?(=)(?=\s*?["\'])', bygroup(Name.Attribute, Operator)
        yield from cls.find_strings()
        yield r'&\S*?;', Escape
        yield r'\?>', Bracket.Preprocessed.End, -1
        yield default_action, Text.Preprocessed

    @lexicon
    def doctype(cls):
        yield words(("SYSTEM", "PUBLIC", "NDATA")), Keyword
        yield _N_, Name
        yield from cls.common_defs()
        yield r'\[', Bracket, cls.internal_dtd
        yield r'>', Delimiter, -1

    @lexicon
    def internal_dtd(cls):
        yield r'\]', Bracket, -1
        yield from Dtd.root

    @lexicon
    def attrs(cls):
        yield _N_, Name.Attribute
        yield r'=', Operator
        yield from cls.find_strings()
        yield r'/\s*>', Delimiter, -1
        yield r'>', Delimiter, -1, cls.tag
        yield r'\s+', skip
        yield default_action, Invalid


class Dtd(_XmlBase):
    """Parse a DTD (Document Type Definition)."""
    @lexicon
    def root(cls):
        yield from cls.find_comments()
        yield fr'(<!)(ENTITY)\b(?:\s*(%))?(?:\s*({_N_}))?', \
            bygroup(Delimiter, Keyword, Keyword, Name.Entity.Definition), cls.entity
        yield fr'(<!)(ELEMENT|ATTLIST|NOTATION)\b(?:\s*({_N_}))?', \
            bygroup(Delimiter, Keyword, Name.Element.Definition), \
            dselect(MATCH(2), {"ELEMENT": cls.element, "ATTLIST": cls.attlist}, cls.notation)
        yield fr'%{_N_};', Name.Entity.Escape
        yield default_action, select(call(str.isspace, TEXT), Text, skip)

    @lexicon
    def entity(cls):
        yield words(("SYSTEM", "PUBLIC", "NDATA")), Keyword
        yield _N_, Name.Entity
        yield from cls.common_defs()
        yield r'>', Delimiter, -1

    @lexicon
    def element(cls):
        yield r'\(', Bracket, cls.element_contents
        yield words(("ANY", "EMPTY")), Name.Keyword
        yield r'[,|?+*]', Operator
        yield from cls.common_defs()
        yield r'>', Delimiter, -1

    @lexicon
    def element_contents(cls):
        """Content definition inside a <!ELEMENT > declaration."""
        yield r'#PCDATA', Name.Builtin
        yield from cls.enumerate(r'[,|?+*]', Name.Element)

    @lexicon
    def attlist(cls):
        yield words(("#REQUIRED", "#IMPLIED", "#FIXED"), suffix=r'\b'), Name.Builtin
        yield words(('CDATA', 'ID', 'IDREF', 'IDREFS', 'ENTITY', 'ENTITIES',
            'NMTOKEN', 'NMTOKENS'), prefix=r'\b', suffix=r'\b'), Name.Type
        yield r'\b(NOTATION)\b(?:\s+(\())', bygroup(Name.Type, Bracket), \
            ifgroup(2, cls.attlist_notation)
        yield _N_, Name.Attribute.Definition
        yield r'\(', Bracket, cls.attlist_enumeration
        yield from cls.common_defs()
        yield r'>', Delimiter, -1

    @lexicon
    def attlist_enumeration(cls):
        yield from cls.enumerate(r'\|', Data)

    @lexicon
    def attlist_notation(cls):
        yield from cls.enumerate(r'\|', Name.Type)

    @lexicon
    def notation(cls):
        yield words(("SYSTEM", "PUBLIC")), Keyword
        yield from cls.common_defs()
        yield r'>', Delimiter, -1

    @classmethod
    def enumerate(cls, operators=r'\|', nametype=Name.Type):
        """Find names between ( ), and operators, string and parameter entities.

        ``operators`` is the regexp for the operators, ``nametype`` the action
        for the found names.

        """
        yield r'\(', Bracket, 1
        yield r'\)', Bracket, -1
        yield operators, Operator
        yield _T_, nametype
        yield from cls.common_defs()



# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
Formatter for HTML output.

This module is in development, and in pre-alpha stage :-)

"""

import parce.formatter


def escape(text):
    """Escape &, < and >."""
    return text.replace('&', "&amp;").replace('<', "&lt;").replace('>', "&gt;")


def attrescape(text):
    """Escape &, <, > and "."""
    return escape(text).replace('"', "&quot;")


def html(tokens, text, theme):
    """Test function to convert text to colored html using theme.

    The text must also be specified because not all text may be covered by
    the tokens.

    """
    f = parce.formatter.Formatter(theme, span_factory_inline)
    yield '<html>'
    yield '<div style="{}"><pre nowrap="nowrap">'.format(attrescape(f.window()))
    oldend = 0
    for pos, end, style, window in f.format_ranges(tokens):
        if pos > oldend:
            yield escape(text[oldend:pos])
        yield '<span style="{}">{}</span>'.format(attrescape(style), escape(text[pos:end]))
        oldend = end
    if oldend < len(text):
        yield escape(text[oldend:])
    yield '</pre></div>'
    yield '</html>'



def span_factory_inline(textformat):
    """Convert a TextFormat to an inline style string."""
    d = textformat.css_properties()
    if d:
        return ' '.join(sorted('{}: {};'.format(prop, value) for prop, value in d.items()))





# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Testing CSS parser with pytest-3.
"""

import sys

sys.path.insert(0, ".")

from parce import root
from parce.action import Name, Number
from parce.lang.css import Css

def test_main():
    css = r"""
h1[attribute="value"] + p {
    width: 500px;
    height: 90%;
    color: white;
    background: url(www.image.org/image.png);
    text-decoration: underline !important;
}
    """

    tree = root(Css.root, css)

    assert len(tree) == 2   # prelude and rule
    assert tree[0][0][2][0][0].action is Name.Attribute
    assert tree.query.all.action(Number).pick().pos == 40
    assert tree.query.all(Css.declaration)[0][0].list() == [
        'width', 'height', 'color', 'background','text-decoration']
    assert tree.query.all.action(Name.Property.Definition)("color").next.next.pick() == "white"

if __name__ == "__main__":
    test_main()



# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Parse Python.

"""

__all__ = ('Python', 'PythonConsole')

import re

from parce import lexicon, Language, skip, default_action, default_target
from parce.action import (
    Character, Comment, Data, Delimiter, Escape, Invalid, Keyword, Literal,
    Name, Number, Operator, String, Whitespace)
from parce.rule import (
    ARG, MATCH, TEXT, arg, bygroup, call, derive, dselect, findmember, ifarg,
    ifeq, ifgroup, ifmember, pattern, select, words)

from . import python_words


RE_PYTHON_IDENTIFIER = _I_ = r'[^\W\d]\w*'
RE_PYTHON_HORIZ_SPACE = _S_ = r'[^\S\n]'
RE_PYTHON_LINE_CONTINUATION = _N_ = r'\\\n'
_SN_ = fr'(?:{_S_}|{_N_})'


Bytes = Data.Bytes


class Python(Language):
    @lexicon(re_flags=re.MULTILINE)
    def root(cls):
        yield fr'^{_S_}+($|(?=#))?', ifgroup(1, Whitespace, Whitespace.Indent)
        yield r'@', Name.Decorator, cls.decorator
        yield fr'(class\b){_S_}*({_I_})', bygroup(Keyword,
            ifmember(MATCH(2), python_words.keywords, Invalid, Name.Class.Definition)), cls.classdef
        yield fr'(def\b){_S_}*({_I_})', bygroup(Keyword,
            ifmember(MATCH(2), python_words.keywords, Invalid, Name.Function.Definition)), cls.funcdef
        yield fr':(?={_S_}*(?:$|#))', Delimiter.Indent
        yield fr'({_I_})\s*(=)', bygroup(
            select(call(str.isupper, TEXT),
                   select(call(isclassname, TEXT), Name.Variable, Name.Class),
                   Name.Constant),
            Operator.Assignment)
        yield from cls.common()

    @classmethod
    def common(cls):
        yield r'#', Comment, cls.comment
        yield fr'({_N_})(\s*)', bygroup(Escape, Whitespace)
        yield r'\[', Delimiter, cls.list
        yield r'\(', Delimiter, cls.tuple
        yield r'\{', Delimiter, cls.dict

        ## string literals
        yield from cls.find_string_literals()
        yield from cls.find_bytes_literals()

        ## numerical values
        yield '0[oO](?:_?[0-7])+', Number
        yield '0[bB](?:_?[01])+', Number
        yield '0[xX](?:_?[0-9a-fA-F])+', Number
        yield r'(?:\.\d(?:_?\d)*|\d(?:_?\d)*(?:\.(?:\d(?:_?\d)*)?)?)(?:[eE][-+]\d(?:_?\d)*)?[jJ]?', Number

        ## keywords, variables, functions
        yield words(python_words.keywords, prefix=r'\b', suffix=r'\b'), Keyword
        yield words(python_words.constants, prefix=r'\b', suffix=r'\b'), Name.Constant
        yield fr'\b(self|cls)\b(?:{_SN_}*([\[\(]))?', Name.Variable.Special, \
            dselect(MATCH(2), {'(': cls.call, '[': cls.item})
        # method, class or attribute (keywords after a . are also caught)
        yield fr'(\.){_SN_}*\b({_I_})\b(?:{_SN_}*([\[\(]))?', \
            bygroup(
                Delimiter,
                ifmember(MATCH(2), python_words.keywords,
                    Keyword,
                    dselect(MATCH(3), {'(': select(call(isclassname, TEXT), Name.Method, Name.Class)},
                         select(call(str.isupper, TEXT),
                             select(call(isclassname, TEXT), Name.Attribute, Name.Class),
                             Name.Constant))),
                Delimiter), \
            dselect(MATCH(3), {'(': cls.call, '[': cls.item})
        # function, class or variable
        yield fr'\b({_I_})\b(?:{_SN_}*([\[\(]))?', \
            bygroup(
                findmember(MATCH(1),
                    ((python_words.builtins, Name.Builtin),
                     (python_words.exceptions, Name.Exception)),
                    select(call(str.isupper, TEXT),
                        select(call(isclassname, TEXT),
                            dselect(MATCH(2), {'(': Name.Function}, Name.Variable),
                            Name.Class),
                        Name.Constant)),
                Delimiter), \
            dselect(MATCH(2), {'(': cls.call, '[': cls.item})

        ## delimiters, operators
        yield r'\.\.\.', Delimiter.Special.Ellipsis
        yield r'(?:\*\*|//|<<|>>|[-+*/%@&|^:])?=', Operator.Assignment
        yield r'\*\*|//|<<|>>|[<>=!]=|[-+*/%@&|^~<>]', Operator
        yield r'[.;,:]', Delimiter

    @lexicon(re_flags=re.MULTILINE)
    def decorator(cls):
        """A decorator."""
        yield _I_, Name.Decorator
        yield r'\[', Delimiter, cls.item
        yield r'\(', Delimiter, cls.call
        yield r'\.', Delimiter
        yield '$', None, -1
        yield r'\\\n', Escape
        yield r'#', Comment, -1, cls.comment

    @lexicon
    def funcdef(cls):
        """A function definition."""
        yield r'\(', Delimiter, cls.signature
        yield r'->', Delimiter.Annotation
        yield r':', Delimiter.Indent, -1
        yield r'#', Comment, -1, cls.comment
        yield from cls.common()

    @lexicon
    def signature(cls):
        """A function signature."""
        yield r'\)', Delimiter, -1
        yield r':', Delimiter.Annotation
        yield from cls.common()

    @lexicon
    def classdef(cls):
        """A class definition."""
        yield r'\(', Delimiter, cls.bases
        yield ":", Delimiter.Indent, -1
        yield r'#', Comment, -1, cls.comment
        yield from cls.common()

    @lexicon
    def bases(cls):
        """The base classes in a class definition."""
        yield r'\)', Delimiter, -1
        yield from cls.common()

    ## ------ expressions -----------
    @lexicon
    def item(cls):
        """Stuff between xxx[ and ] (getitem)."""
        yield r'\]', Delimiter, -1
        yield from cls.common()

    @lexicon
    def call(cls):
        """Stuff between xxx( and ) (call)."""
        yield r'\)', Delimiter, -1
        yield from cls.common()

    ## ----- item types -------------
    @lexicon
    def list(cls):
        yield r'\]', Delimiter, -1
        yield ',', Delimiter
        yield from cls.common()

    @lexicon
    def tuple(cls):
        yield r'\)', Delimiter, -1
        yield ',', Delimiter
        yield from cls.common()

    @lexicon
    def dict(cls):
        yield r'\}', Delimiter, -1
        yield '[,:]', Delimiter
        yield from cls.common()

    ## ------- strings --------------
    @classmethod
    def find_string_literals(cls, target=None, allow_newlines=None):
        """Find string literals."""
        # short strings not closed on the same line are invalid
        yield r'''[rRuUfF]{,2}["']$''', String.Invalid

        if target is None:
            target = cls.string(allow_newlines)

        # long strings
        yield r'(\b[rR])("""|'r"''')", \
            bygroup(String.Prefix, String.Start), \
            target, derive(cls.long_string_raw, MATCH(2))
        yield r'(\b(?:[fF][rR])|(?:[rR][fF]))("""|'r"''')", \
            bygroup(String.Prefix, String.Start), \
            target, derive(cls.long_string_raw_format, MATCH(2))
        yield r'(\b[uU])?("""|'r"''')", \
            bygroup(String.Prefix, String.Start), \
            target, derive(cls.long_string, MATCH(2))
        yield r'(\b[fF])("""|'r"''')", \
            bygroup(String.Prefix, String.Start), \
            target, derive(cls.long_string_format, MATCH(2))

        # short strings
        yield r'''(\b[rR])(['"])''', \
            bygroup(String.Prefix, String.Start), \
            target, derive(cls.short_string_raw, MATCH(2))
        yield r'''(\b(?:[fF][rR])|(?:[rR][fF]))(['"])''', \
            bygroup(String.Prefix, String.Start), \
            target, derive(cls.short_string_raw_format, MATCH(2))
        yield r'''(\b[uU])?(['"])''', \
            bygroup(String.Prefix, String.Start), \
            target, derive(cls.short_string, MATCH(2))
        yield r'''(\b[fF])(['"])''', \
            bygroup(String.Prefix, String.Start), \
            target, derive(cls.short_string_format, MATCH(2))

    @lexicon
    def string(cls):
        """All strings end here, check [slice] notation and concatenated literals."""
        yield _N_, Escape
        yield ifarg(r'\s+', r'[ \t]+'), skip    # allow newline inside arglists, tuples, etc
        yield from cls.find_string_literals(0)
        yield r'\[', Delimiter, cls.item
        yield default_target, -1

    @lexicon(re_flags=re.MULTILINE)
    def short_string(cls):
        yield from cls.string_escape()
        yield from cls.short_string_common()

    @lexicon(re_flags=re.MULTILINE)
    def short_string_raw(cls):
        yield from cls.short_string_raw_common()

    @lexicon(re_flags=re.MULTILINE)
    def short_string_format(cls):
        yield from cls.string_formatstring()
        yield from cls.string_escape()
        yield from cls.short_string_common()

    @lexicon(re_flags=re.MULTILINE)
    def short_string_raw_format(cls):
        yield from cls.string_formatstring()
        yield from cls.short_string_raw_common()

    @classmethod
    def short_string_common(cls):
        yield arg(), String.End, -1
        yield pattern(ifeq(ARG, "'", r"[^']*?$", r'[^"]*?$')), String.Invalid, -1
        yield default_action, String

    @classmethod
    def short_string_raw_common(cls):
        yield arg(), String.End, -1
        yield r'\\\\', String
        yield pattern(ifeq(ARG, "'", fr"([^\\']*?|\\'{_S_}*)$", fr'([^\\"]*?|\\"{_S_}*)$')), String.Invalid, -1
        yield arg(prefix=r'\\'), String  # escape quote, but the \ remains
        yield default_action, String

    @lexicon
    def long_string(cls):
        yield from cls.string_escape()
        yield from cls.long_string_common()

    @lexicon
    def long_string_raw(cls):
        yield arg(prefix=r'\\'), String  # escape quote, but the \ remains
        yield from cls.long_string_common()

    @lexicon
    def long_string_format(cls):
        yield from cls.string_formatstring()
        yield from cls.string_escape()
        yield from cls.long_string_common()

    @lexicon
    def long_string_raw_format(cls):
        yield arg(prefix=r'\\'), String  # escape quote, but the \ remains
        yield from cls.string_formatstring()
        yield from cls.long_string_common()

    @classmethod
    def long_string_common(cls):
        yield arg(), String.End, -1
        yield default_action, String

    # ------ stuff common for short and long strings ---------
    @classmethod
    def string_escape(cls):
        yield from cls.bytes_escape(String.Escape)
        yield r'\\N\{[^\}]+\}', String.Escape
        yield r'\\u[0-9a-fA-F]{4}', String.Escape
        yield r'\\U[0-9a-fA-F]{8}', String.Escape

    @classmethod
    def string_formatstring(cls):
        yield r'\{\{|\}\}', String.Escape
        yield r'\{', Delimiter.Template, cls.string_format_expr

    @lexicon
    def string_format_expr(cls):
        yield '![sra]', Character
        yield ':', Delimiter, cls.string_format_spec
        yield r'\}', Delimiter.Template, -1
        yield from cls.common()

    @lexicon
    def string_format_spec(cls):
        yield r'\{', Delimiter, cls.string_format_spec_nested
        yield r'\}', Delimiter.Template, -2
        yield from cls.common() # TODO maybe really parse format strings

    @lexicon
    def string_format_spec_nested(cls):
        yield r'\}', Delimiter, -1
        yield from cls.common()

    # ----------------- bytes --------------------
    @classmethod
    def find_bytes_literals(cls, target=None, allow_newlines=None):
        """Find bytes literals."""
        # short bytes not closed on the same line are invalid
        yield r'''[rRbB]{,2}["']$''', Bytes.Invalid

        if target is None:
            target = cls.bytes(allow_newlines)

        # long bytes
        yield r'(\b(?:[bB][rR])|(?:[rR][bB]))("""|'r"''')", \
            bygroup(Bytes.Prefix, Bytes.Start), \
            target, derive(cls.long_bytes_raw, MATCH(2))
        yield r'(\b[bB])("""|'r"''')", \
            bygroup(Bytes.Prefix, Bytes.Start), \
            target, derive(cls.long_bytes, MATCH(2))

        # short bytes
        yield r'''(\b(?:[bB][rR])|(?:[rR][bB]))(['"])''', \
            bygroup(Bytes.Prefix, Bytes.Start), \
            target, derive(cls.short_bytes_raw, MATCH(2))
        yield r'''(\b[bB])(['"])''', \
            bygroup(Bytes.Prefix, Bytes.Start), \
            target, derive(cls.short_bytes, MATCH(2))

    @lexicon
    def bytes(cls):
        """All bytes end here, check [slice] notation and concatenated literals."""
        yield _N_, Escape
        yield ifarg(r'\s+', r'[ \t]+'), skip    # allow newline inside arglists, tuples, etc
        yield from cls.find_bytes_literals(0)
        yield r'\[', Delimiter, cls.item
        yield default_target, -1

    @lexicon(re_flags=re.MULTILINE)
    def short_bytes(cls):
        yield from cls.bytes_escape()
        yield from cls.short_bytes_common()

    @lexicon(re_flags=re.MULTILINE)
    def short_bytes_raw(cls):
        yield from cls.short_bytes_raw_common()

    @lexicon
    def long_bytes(cls):
        yield from cls.bytes_escape()
        yield from cls.long_bytes_common()

    @lexicon
    def long_bytes_raw(cls):
        yield from cls.long_bytes_common()

    @classmethod
    def short_bytes_common(cls):
        yield arg(), Bytes.End, -1
        yield pattern(ifeq(ARG, "'", r"[^']*?$", r'[^"]*?$')), Bytes.Invalid, -1
        yield default_action, Bytes

    @classmethod
    def short_bytes_raw_common(cls):
        yield r'\\\\', Bytes
        yield pattern(ifeq(ARG, "'", fr"([^\\']*?|\\'{_S_}*)$", fr'([^\\"]*?|\\"{_S_}*)$')), Bytes.Invalid, -1
        yield arg(prefix=r'\\'), Bytes  # escape quote, but the \ remains
        yield from cls.long_bytes_common()

    @classmethod
    def long_bytes_common(cls):
        yield arg(), Bytes.End, -1
        yield default_action, Bytes

    @classmethod
    def bytes_escape(cls, action=Bytes.Escape):
        yield r'''\\[\n\\'"abfnrtv]''', action
        yield r'\\\d{1,3}', action
        yield r'\\x[0-9a-fA-F]{2}', action

    ## ------- comments -------------
    @lexicon(re_flags=re.MULTILINE)
    def comment(cls):
        yield from cls.comment_common()
        yield r'$', Comment, -1


class PythonConsole(Python):
    """Python console input and output with prompt."""
    @classmethod
    def common(cls):
        yield r'(?:(?<=\n)|^)(?:>>>|\.\.\.) ', Literal.Prompt
        yield from super().common()

    @classmethod
    def longstring_common(cls):
        yield r'(?:(?<=\n)|^)\.\.\. ', Literal.Prompt
        yield from super().longstring_common()

    @classmethod
    def longbytes_common(cls):
        yield r'(?:(?<=\n)|^)\.\.\. ', Literal.Prompt
        yield from super().longbytes_common()


def isclassname(text):
    """Return True if text starts with uppercase letter.

    Starting underscores are skipped.

    """
    for c in text:
        if c.isupper():
            return True
        elif c != '_':
            return False
    return False





# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
JavaScript Object Notation parser.

Numbers become Number tokens, ``true``, ``false`` and ``null`` become
Name.Constant tokens, strings are yielded in ``string`` contexts, with String
tokens and possibily String.Escape tokens for escaped characters.

Objects (``{ ... }``) become ``object`` contexts with alternating ``key`` and
``value`` child contexts. Arrays (``[ ... ]``) become ``array`` contexts.

"""

__all__ = ('Json', 'JsonTransform')

import re

from parce import Language, lexicon, skip, default_action, default_target
from parce.action import Delimiter, Name, Number, String
from parce.rule import words
from parce.transform import Transform


JSON_CONSTANTS = {
    'true': True,
    'false': False,
    'null': None,
}


class Json(Language):
    @lexicon
    def root(cls):
        yield from cls.values()

    @classmethod
    def values(cls):
        yield r"\{", Delimiter, cls.object
        yield r"\[", Delimiter, cls.array
        yield '"', String, cls.string
        yield r"-?\d+(?:\.\d+)?(?:[Ee][+-]?\d+)?", Number
        yield words(JSON_CONSTANTS, r'\b', r'\b'), Name.Constant

    @lexicon
    def object(cls):
        yield r"\}", Delimiter, -1
        yield r"\s+", skip
        yield default_target, cls.key

    @lexicon
    def key(cls):
        yield '"', String, cls.string
        yield ":", Delimiter, -1, cls.value

    @lexicon
    def value(cls):
        yield from cls.values()
        yield ",", Delimiter, -1
        yield r"\}", Delimiter, -2

    @lexicon
    def array(cls):
        yield from cls.values()
        yield ",", Delimiter
        yield r"\]", Delimiter, -1

    @lexicon
    def string(cls):
        yield '"', String, -1
        yield r'\\(?:["\\/bfnrt]|u[0-9a-fA-F]{4})', String.Escape
        yield default_action, String


class JsonTransform(Transform):
    """Transforms a Json expression tree to the Python equivalent."""
    def root(self, items):
        for value in self.values(items):
            return value

    def values(self, items):
        """Yield values like the Json.values() classmethod generates."""
        for i in items:
            if i.is_token:
                if i.action == Number:
                    n = float(i.text)
                    if n.is_integer():
                        n = int(n)
                    yield n
                elif i.action == Name.Constant:
                    yield JSON_CONSTANTS[i.text]
            else:
                yield i.obj

    def object(self, items):
        d = {}
        key = None
        for name, obj in items.items("key", "value"):
            if name == "key":
                key = obj
            elif key is not None:
                d[key] = obj
                key = None
        return d

    def key(self, items):
        for i in items.items("string"):
            return i.obj

    def value(self, items):
        for value in self.values(items):
            return value

    def array(self, items):
        return list(self.values(items))

    def string(self, items):
        def gen():
            for i in items[:-1].tokens():
                if i.action == String.Escape:
                    if i.text[1] == 'u':
                        yield chr(int(i.text[2:], 16))
                    else:
                        yield i.text[1]
                else:
                    yield i.text
        return ''.join(gen())



# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
This module defines the tree structure a text is parsed into.

A tree consists of Context and Token objects.

A Context is a list containing Tokens and other Contexts. A Context is created
when a lexicon becomes active. A Context knows its parent Context and its
lexicon.

A Token represents one parsed piece of text. A Token is created when a rule in
the lexicon matches. A Token knows its parent Context, its position in the text
and the action that was specified in the rule.

The root Context is always one, and it represents the root lexicon. A Context
is always non-empty, except for the root Context, which is empty if the
document did not generate a single token.

The tree structure is very easy to navigate, no special objects or iterators
are necessary for that.

To find a token at a certain position in a context, use find_token() and its
relatives. From every token you can iterate forward() and backward(). Use the
methods like left_siblings() and right_siblings() to traverse the current
context.

See also the documentation for Token and Context.

"""


import itertools
import reprlib
import weakref

from parce import util
from parce import query
from parce.lexicon import Lexicon


DUMP_STYLES = {
    "ascii":   (" | ", "   ", " |-", " `-"),
    "round":   (" │ ", "   ", " ├╴", " ╰╴"),
    "square":  (" │ ", "   ", " ├╴", " └╴"),
    "double":  (" ║ ", "   ", " ╠═", " ╚═"),
    "thick":   (" ┃ ", "   ", " ┣╸", " ┗╸"),
    "flat":    ("│", " ", "├", "╰"),
}

DUMP_STYLE_DEFAULT = "round"


class Node:
    """Methods that are shared by Token and Context."""
    __slots__ = ('__weakref__',)

    is_token = False
    is_context = False

    @property
    def parent(self):
        """The parent Context (or None; uses a weak reference)."""
        return self._parent()

    @parent.setter
    def parent(self, parent):
        """Set the parent (to a Context or None)."""
        self._parent = weakref.ref(parent) if parent is not None else lambda: None

    @parent.deleter
    def parent(self):
        """Set the parent to None."""
        self._parent = lambda: None

    def dump(self, file=None, style=None, depth=0):
        """Display a graphical representation of the node and its contents.

        The file object defaults to stdout, and the style to "round". You can
        choose any style that's in the ``DUMP_STYLES`` dictionary.

        """
        i = 2
        d = DUMP_STYLES[style or DUMP_STYLE_DEFAULT]
        prefix = []
        node = self
        for _ in range(depth):
            prefix.append(d[i + int(node.is_last())])
            node = node.parent
            i = 0
        print("".join(reversed(prefix)) + repr(self), file=file)
        if self.is_context:
            for n in self:
                n.dump(file, style, depth + 1)

    def parent_index(self):
        """Return our index in the parent.

        This is recommended above using parent.index(self), because this method
        finds our index using a binary search on position, while the latter
        is a linear search, which is certainly slower with a large number of
        children.

        """
        p = self.parent
        pos = self.pos
        lo = 0
        hi = len(p)
        while lo < hi:
            mid = (lo + hi) // 2
            n = p[mid]
            if n.pos < pos:
                lo = mid + 1
            elif n is self:
                return mid
            else:
                hi = mid
        return lo

    def root(self):
        """Return the root node."""
        root = self
        for root in self.ancestors():
            pass
        return root

    def is_root(self):
        """Return True if this Node has no parent node."""
        return self.parent is None

    def is_last(self):
        """Return True if this Node is the last child of its parent.

        Fails if called on the root element.

        """
        return self.parent[-1] is self

    def is_first(self):
        """Return True if this Node is the first child of its parent.

        Fails if called on the root element.

        """
        return self.parent[0] is self

    def is_ancestor_of(self, node):
        """Return True if this Node is an ancestor of the other Node."""
        for n in node.ancestors():
            if n is self:
                return True
        return False

    def ancestors(self, upto=None):
        """Climb the tree up over the parents.

        If upto is given and it is one of the ancestors, stop after yielding
        that ancestor. Otherwise iteration stops at the root node.

        """
        node = self.parent
        if upto and upto.parent is not None:
            p = upto.parent
            while node is not None and node is not p:
                yield node
                node = node.parent
        else:
            while node is not None:
                yield node
                node = node.parent

    def ancestors_with_index(self, upto=None):
        """Yield the ancestors(upto), and the index of each node in the parent."""
        n = self
        for p in self.ancestors(upto):
            yield p, n.parent_index()
            n = p

    def common_ancestor(self, other):
        """Return the common ancestor with the Context or Token."""
        ancestors = []
        if self.is_context:
            ancestors.append(self)
        ancestors.extend(self.ancestors())
        if other.is_context and other in ancestors:
            return other
        for n in other.ancestors():
            if n in ancestors:
                return n

    def depth(self):
        """Return the number of ancestors."""
        return sum(1 for n in self.ancestors())

    def left_sibling(self):
        """Return the left sibling of this node, if any.

        Does not descend in child nodes or ascend upto the parent.
        Fails if called on the root node.

        """
        if self.parent[0] is not self:
            i = self.parent_index()
            return self.parent[i-1]

    def right_sibling(self):
        """Return the right sibling of this node, if any.

        Does not descend in child nodes or ascend upto the parent.
        Fails if called on the root node.

        """
        if self.parent[-1] is not self:
            i = self.parent_index()
            return self.parent[i+1]

    def left_siblings(self):
        """Yield the left siblings of this node in reverse order, if any.

        Does not descend in child nodes or ascend upto the parent.
        Fails if called on the root node.

        """
        if self.parent[0] is not self:
            i = self.parent_index()
            yield from self.parent[i-1::-1]

    def right_siblings(self):
        """Yield the right siblings of this node, if any.

        Does not descend in child nodes or ascend upto the parent.
        Fails if called on the root node.

        """
        if self.parent[-1] is not self:
            i = self.parent_index()
            yield from self.parent[i+1:]

    def next_token(self):
        """Return the following Token, if any."""
        for t in self.forward():
            return t

    def previous_token(self):
        """Return the preceding Token, if any."""
        for t in self.backward():
            return t

    def forward(self, upto=None):
        """Yield all Tokens in forward direction.

        Descends into child Contexts, and ascends into parent Contexts.
        If upto is given, does not ascend above that context.

        """
        for parent, index in self.ancestors_with_index(upto):
            yield from util.tokens(parent[index+1:])

    def backward(self, upto=None):
        """Yield all Tokens in backward direction.

        Descends into child Contexts, and ascends into parent Contexts.
        If upto is given, does not ascend above that context.

        """
        for parent, index in self.ancestors_with_index(upto):
            if index:
                yield from util.tokens_bw(parent[index-1::-1])

    @property
    def query(self):
        """Query this node in different ways; see the :mod:`~parce.query` module."""
        def gen():
            yield self
        return query.Query(gen)

    def delete(self):
        """Remove this node from its parent.

        If the parent would become empty, it is removed too.
        Returns the first non-empty ancestor.

        """
        for parent, index in self.ancestors_with_index():
            del parent[index]
            if len(parent):
                return parent


class Token(Node):
    """A Token instance represents a lexed piece of text.

    A token has the following attributes:

    `parent`:
        the Context node to which the token was added
    `pos`:
        the position of the token in the original text
    `end`:
        the end position of the token in the original text
    `text`:
        the text of the token
    `action`:
        the action specified by the lexicon rule that created the token

    When a pattern rule in a lexicon matches the text, a Token is created. When
    that rule would create more than one Token from a single regular expression
    match, _GroupToken objects are created instead, carrying the index of the
    token in the group in the `group` attribute. The `group` attribute is
    readonly None for normal tokens.

    GroupTokens are thus always adjacent in the same context. If you want to
    retokenize text starting at some position, be sure you are at the start of
    a grouped token, e.g.::

        t = ctx.find_token(45)
        if t.group:
            for t in t.left_siblings():
                if not t.group:
                    break
        pos = t.pos

    (A _GroupToken is just a normal Token otherwise, the reason a subclass was
    created is that the group attribute is unused in by far the most tokens, so
    it does not use any memory. You never need to reference the _GroupToken
    class; just test the group attribute if you want to know if a token belongs
    to a group that originated from a single match.)

    When iterating over the children of a Context (which may be Context or
    Token instances), you can use the `is_token` attribute to determine whether
    the node child is a token, which is easier than to call `isinstance(t,
    Token)` each time.

    From a token, you can iterate `forward()` or `backward()` to find adjacent
    tokens. If you only want to stay in the current context, use the various
    sibling methods, such as `right_sibling()`.

    By traversing the `ancestors()` of a token or context, you can find which
    lexicons created the tokens.

    You can compare a Token instance with a string. Instead of::

        if token.text == "bla":
            do_something()

    you can do::

        if token == "bla":
            do_something()

    You can call `len()` on a token, which returns the length of the token's
    text attribute, and you can use the string format method to embed the
    token's text in another string::

        s = "blabla {}".format(token)

    A token always has a parent, and that parent is always a Context instance.

    """

    __slots__ = "_parent", "pos", "text", "action"

    is_token = True
    group = None

    def __init__(self, parent, pos, text, action):
        self.parent = parent
        self.pos = pos
        self.text = text
        self.action = action

    def copy(self, parent=None):
        """Return a copy of the Token, but with the specified parent."""
        return type(self)(parent, self.pos, self.text, self.action)

    def equals(self, other):
        """Return True if the other Token has the same ``text`` and ``action``
        attributes and the same context ancestry (see also
        :meth:`state_matches`).

        Note that the ``pos`` attribute is not compared.

        """
        return (self.text == other.text
                and self.action == other.action
                and self.state_matches(other))

    def state_matches(self, other):
        """Return True if the other Token has the same lexicons in the ancestors."""
        if other is self:
            return True
        for c1, c2 in zip(self.ancestors(), other.ancestors()):
            if c1 is c2:
                return True
            elif c1.lexicon != c2.lexicon:
                return False
        return c1.parent is None and c2.parent is None

    def __repr__(self):
        text = reprlib.repr(self.text)
        return "<Token {} at {}:{} ({})>".format(text, self.pos, self.end, self.action)

    def __hash__(self):
        return Node.__hash__(self)

    def __eq__(self, other):
        if isinstance(other, str):
            return other == self.text
        return other is self

    def __ne__(self, other):
        if isinstance(other, str):
            return other != self.text
        return other is not self

    def __format__(self, formatstr):
        return self.text.__format__(formatstr)

    def __len__(self):
        return len(self.text)

    @property
    def end(self):
        return self.pos + len(self.text)

    def forward_including(self, upto=None):
        """Yield all tokens in forward direction, including self."""
        yield self
        yield from self.forward(upto)

    def backward_including(self, upto=None):
        """Yield all tokens in backward direction, including self."""
        yield self
        yield from self.backward(upto)

    def forward_until_including(self, other):
        """Yield all tokens starting with us and upto and including the other."""
        context, start_trail, end_trail = self.common_ancestor_with_trail(other)
        if context:
            for context, slice_ in context.slices(start_trail, end_trail):
                yield from util.tokens(context[slice_])

    def common_ancestor_with_trail(self, other):
        """Return a three-tuple(context, trail_self, trail_other).

        The context is the common ancestor such as returned by common_ancestor,
        if any. trail_self is a tuple of indices from the common ancestor upto
        self, and trail_other is a tuple of indices from the same ancestor upto
        the other Token.

        If there is no common ancestor, all three are None. But normally,
        all nodes share the root context, so that will normally be the upmost
        common ancestor.

        """
        if other is self:
            i = self.parent_index()
            return self.parent, (i,), (i,)
        if other.pos > self.pos:
            s_ancestors, s_indices = zip(*self.ancestors_with_index())
            o_indices = []
            for n, i in other.ancestors_with_index():
                o_indices.append(i)
                try:
                    s_i = s_ancestors.index(n)
                except ValueError:
                    continue
                return n, s_indices[s_i::-1], o_indices[::-1]
        return None, None, None

    def target(self):
        """Return the first context directly to the right of this Token.

        The context should be the right sibling of the token, or of any of its
        ancestors. If the token is part of a group, the context is found
        immediately next to the last member of the group. The found context may
        also be a child of the grand-parents of this token, in case the target
        popped contexts first.

        In all cases, the returned context is the one started by the target
        in the lexicon rule that created this token.

        """
        node = self
        if node.group is not None:
            # go to the last one of this group
            group = node.group + 1
            for n in node.right_siblings():
                if n.is_context or n.group != group:
                    break
                node = n
                group += 1
        while node.parent:
            r = node.right_sibling()
            if r:
                if r.is_context:
                    return r
                return
            node = node.parent


class _GroupToken(Token):
    """A Token class that allows setting the `group` attribute."""
    __slots__ = "group",

    def __init__(self, group, parent, pos, text, action):
        self.group = group
        super().__init__(parent, pos, text, action)

    def copy(self, parent=None):
        """Return a copy of the Token, but with the specified parent."""
        return type(self)(self.group, parent, self.pos, self.text, self.action)


class Context(list, Node):
    """A Context represents a list of tokens and contexts.

    The lexicon that created the tokens is in the `lexicon` attribute.

    If a pattern rule jumps to another lexicon, a sub-Context is created and
    tokens are added there. If that lexicon pops back to the current one, new
    tokens can appear after the sub-context. (So the token that caused the jump
    to the sub-context normally preceeds the context it created.)

    A context has a `parent` attribute, which can point to an enclosing
    context. The root context has `parent` None.

    When iterating over the children of a Context (which may be Context or
    Token instances), you can use the `is_context` attribute to determine
    whether the node child is a context, which is easier than to call
    `isinstance(node, Context)` each time.

    You can quickly find tokens in a context, based on text::

        if "bla" in context:
            # etc

    Or child contexts, based on lexicon::

        if MyLanguage.lexicon in context:
            # etc

    And if you want to know which token is on a certain position in the text,
    use e.g.::

        context.find_token(45)

    which, using a bisection algorithm, quickly returns the token, which
    might be in any sub-context of the current context.

    """
    __slots__ = "lexicon", "_parent"

    is_context = True

    def __new__(cls, lexicon, parent):
        return list.__new__(cls)

    def __init__(self, lexicon, parent):
        self.lexicon = lexicon
        self.parent = parent

    def __repr__(self):
        pos, end = self.pos, self.end
        if pos == end:
            pos = end = "?" # both are 0 in this case: empty Context
        name = self.lexicon and repr(self.lexicon)
        children = "child" if len(self) == 1 else "children"
        return "<Context {} at {}-{} ({} {})>".format(
            name, pos, end, len(self), children)

    def __hash__(self):
        return Node.__hash__(self)

    def __eq__(self, other):
        if isinstance(other, Lexicon):
            return other.equals(self.lexicon)
        return other is self

    def __ne__(self, other):
        if isinstance(other, Lexicon):
            return not other.equals(self.lexicon)
        return other is not self

    def copy(self, parent=None):
        """Return a copy of the context, but with the specified parent."""
        # a non-recursive implementation due to Python's recursion limits
        copy = copy_root = type(self)(self.lexicon, parent)
        stack = []
        i = 0
        n = self
        while True:
            z = len(n)
            while i < z:
                m = n[i]
                if m.is_context:
                    copy.append(type(m)(m.lexicon, copy))
                    copy = copy[-1]
                    stack.append(i)
                    i = 0
                    n = m
                    break
                elif m.group is not None:
                    for g, j in enumerate(range(i + 1, z), m.group + 1):
                        if n[j].is_context or n[j].group != g:
                            copy.extend(node.copy(copy) for node in n[i:j])
                            i = j
                            break
                    else:
                        copy.append(m.copy(copy))
                        i += 1
                else:
                    copy.append(m.copy(copy))
                    i += 1
            else:
                if stack:
                    copy = copy.parent
                    n = n.parent
                    i = stack.pop() + 1
                else:
                    break
        return copy_root

    @property
    def pos(self):
        """Return the position or our first token. Returns 0 if empty."""
        try:
            node = self[0]
            while node.is_context:
                node = node[0]
            return node.pos
        except IndexError:
            return 0

    @property
    def end(self):
        """Return the end position or our last token. Returns 0 if empty."""
        try:
            node = self[-1]
            while node.is_context:
                node = node[-1]
            return node.end
        except IndexError:
            return 0

    def height(self):
        """Return the height of the tree (the longest distance to a descendant)."""
        if not self:
            return 0
        stack = []
        height = 0
        i = 0
        n = self
        while True:
            for i in range(i, len(n)):
                m = n[i]
                if m.is_context:
                    stack.append(i)
                    height = max(height, len(stack))
                    i = 0
                    n = m
                    break
            else:
                if stack:
                    n = n.parent
                    i = stack.pop() + 1
                else:
                    return height + 1

        # originally, but that led to recursion errors...
        #return max(n.height() + 1 if n.is_context else 1 for n in self) if self else 0

    def tokens(self):
        """Yield all Tokens, descending into nested Contexts."""
        stack = []
        i = 0
        n = self
        while True:
            for i in range(i, len(n)):
                m = n[i]
                if m.is_token:
                    yield m
                else:
                    stack.append(i)
                    i = 0
                    n = m
                    break
            else:
                if stack:
                    n = n.parent
                    i = stack.pop() + 1
                else:
                    break

    def tokens_bw(self):
        """Yield all Tokens, descending into nested Contexts, in backward direction."""
        stack = []
        n = self
        i = len(n)
        while True:
            for i in range(i - 1, -1, -1):
                m = n[i]
                if m.is_token:
                    yield m
                else:
                    stack.append(i)
                    n = m
                    i = len(n)
                    break
            else:
                if stack:
                    n = n.parent
                    i = stack.pop()
                else:
                    break

    def first_token(self):
        """Return our first Token."""
        try:
            node = self[0]
            while node.is_context:
                node = node[0]
            return node
        except IndexError:
            pass

    def last_token(self):
        """Return our last token."""
        try:
            node = self[-1]
            while node.is_context:
                node = node[-1]
            return node
        except IndexError:
            pass

    def find(self, pos):
        """Return the index of our child at pos."""
        i = 0
        hi = len(self)
        while i < hi:
            mid = (i + hi) // 2
            n = self[mid]
            if n.pos >= pos:
                hi = mid
            elif n.end <= pos:
                i = mid + 1
            else:
                hi = mid
        return min(i, len(self) - 1)

    def find_token(self, pos):
        """Return the Token at or to the right of position."""
        n = self[self.find(pos)]
        while n.is_context:
            n = n[n.find(pos)]
        return n

    def find_token_with_trail(self, pos):
        """Return the Token at or to the right of position, and the trail of indices."""
        i = self.find(pos)
        n = self[i]
        trail = [i]
        while n.is_context:
            i = n.find(pos)
            n = n[i]
            trail.append(i)
        return n, trail

    def find_left(self, pos):
        """Return the index of our child at or to the left of pos."""
        i = 0
        hi = len(self)
        while i < hi:
            mid = (i + hi) // 2
            n = self[mid]
            if n.pos < pos:
                i = mid + 1
            else:
                hi = mid
        return max(0, i - 1)

    def find_token_left(self, pos):
        """Return the Token at or to the left of position."""
        n = self[self.find_left(pos)]
        while n.is_context:
            n = n[n.find_left(pos)]
        return n

    def find_token_left_with_trail(self, pos):
        """Return the Token at or to the left of position, and the trail of indices."""
        i = self.find_left(pos)
        n = self[i]
        trail = [i]
        while n.is_context:
            i = n.find_left(pos)
            n = n[i]
            trail.append(i)
        return n, trail

    def find_token_after(self, pos):
        """Return the first token completely right from pos.

        Returns None if there is no token right from pos.

        """
        node = self
        while True:
            i = 0
            hi = l = len(node)
            while i < hi:
                mid = (i + hi) // 2
                n = node[mid]
                if n.is_context:
                    n = n.last_token()
                if n.pos < pos:
                    i = mid + 1
                else:
                    hi = mid
            if i >= l:
                return
            node = node[i]
            if node.is_token:
                return node

    def find_token_before(self, pos):
        """Return the last token completely left from pos.

        Returns None if there is no token left from pos.

        """
        node = self
        while True:
            i = 0
            hi = len(node)
            while i < hi:
                mid = (i + hi) // 2
                n = node[mid]
                if n.is_context:
                    n = n.first_token()
                if pos < n.end:
                    hi = mid
                else:
                    i = mid + 1
            if i == 0:
                return
            node = node[i-1]
            if node.is_token:
                return node

    def tokens_range(self, start=0, end=None):
        """Yield all tokens (that completely fill this text range if specified).

        The first and last tokens may overlap with the start and end positions.

        """
        for context, slice_ in self.context_slices(start, end):
            yield from util.tokens(context[slice_])

    def context_slices(self, start=0, end=None):
        """Yield (context, slice) tuples to yield tokens from.

        Yield the tokens using the context[slice] notation. The first and
        last tokens that would be yielded from the iterables may overlap with
        the start and end positions.

        """
        context, start_trail, end_trail = self.context_trails(start, end)
        if context:
            yield from context.slices(start_trail, end_trail)

    def context_trails(self, start=0, end=None):
        """Return a three-tuple(context, start_trail, end_trail).

        This can be used to denote a range of the tree structure in slices. The
        returned context is the common ancestor of the tokens found at start
        and end (or the current node if start or end fall outside the range of
        the node). The trails are (possibly empty) lists of indices pointing to
        the start and end token, if any.

        """
        if not self:
            return None, None, None  # empty
        context = self
        if end is not None and end < self.end:
            if end <= start:
                return None, None, None
            end_trail = self.find_token_left_with_trail(end)[1]
        else:
            end_trail = []
        if start > 0:
            start_trail = self.find_token_with_trail(start)[1]
            if end_trail:
                # find the youngest common ancestor
                for n, (i, j) in enumerate(zip(start_trail, end_trail)):
                    if i != j or context[i].is_token:
                        break
                    context = context[i]
                if n:
                    del start_trail[:n]
                    del end_trail[:n]
        else:
            start_trail = []
        return context, start_trail, end_trail

    def slices(self, start_trail, end_trail, target_factory=None):
        """Yield from the current context (context, slice) tuples.

        ``start_trail`` and ``end_trail`` both are lists of indices that
        point to descendant tokens of this context. The yielded slices
        include these tokens.

        If you specify a ``target_factory``, it should be a TargetFactory
        object, and it will be updated along with the yielded slices.

        """
        if start_trail:
            start = start_trail[0]
            if len(start_trail) > 1:
                ancestors = []
                n = self[start]
                for i in start_trail[1:]:
                    ancestors.append((n, i))
                    n = n[i]
                yield ancestors[-1][0], slice(i, None) # include start token
                for p, i in ancestors[-2::-1]:
                    target_factory and target_factory.pop()
                    yield p, slice(i + 1, None)
                target_factory and target_factory.pop()
                start += 1
        else:
            start = 0
        if end_trail:
            end = end_trail[0]
            if len(end_trail) == 1:
                yield self, slice(start, end + 1)    # include end token
            else:
                yield self, slice(start, end)
                n = self[end]
                for end in end_trail[1:-1]:
                    target_factory and target_factory.push(n.lexicon)
                    yield n, slice(end)
                    n = n[end]
                target_factory and target_factory.push(n.lexicon)
                yield n, slice(end_trail[-1] + 1)   # include end token
        else:
            yield self, slice(start, None)

    def source(self):
        """Return the first Token, if any, when going to the left from this context.

        The returned token is the one that created us, that this context the
        target is for. If the token is member of a group, the first group member
        is returned.

        """
        for token in self.backward():
            if not token.group:
                return token


def make_tokens(event, parent=None):
    """Factory returning a tuple of one or more Token instances for the event.

    The event is an Event namedtuple defined in the mod:`~parce.lexer` module.
    If the event contains more than one token, _GroupToken instances are
    created.

    """
    if len(event.tokens) > 1:
        return tuple(_GroupToken(n, parent, *t) for n, t in enumerate(event.tokens))
    else:
        return Token(parent, *event.tokens[0]),


def get_group(token):
    """For a token that belongs to a group, return the whole group as a list."""
    p = token.parent
    i = j = token.parent_index()
    while i and p[i].group > 0 and p[i-1].is_token and p[i-1].group is not None and p[i-1].group < p[i].group:
        i -= 1
    z = len(p)
    j += 1
    while j < z and p[j].is_token and p[j].group and p[j].group > p[j-1].group:
        j += 1
    return p[i:j]




# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
Updates the language stub files in source/lang and the listing in source/langs.inc
"""

EXAMPLES_DIRECTORY = "../tests/lang/"


import collections
import glob
import io
import os
import sys

sys.path.insert(0, "..")

import parce
from parce.language import get_all_modules, get_languages

STUB = r"""{0}
{1}

.. automodule:: parce.lang.{0}
    :members:
    :undoc-members:
    :show-inheritance:

"""

EXAMPLE_STUB = r"""

Root lexicon ``{0}`` and text:

.. code-block:: {1}

{2}

Result tree::

{3}


"""

LANGS_INC_HEADER = r"""
.. Generated by ../update_languages.py -- do not edit!

.. list-table::
   :header-rows: 1
   :widths: 20 70

   * - Module
     - Languages

"""

def indent(text, indent=4):
    """Indent text."""
    return '\n'.join(' ' * indent + line for line in text.splitlines())


def make_stub(name, examples):
    text = STUB.format(name, '=' * len(name))
    with open("source/lang/{0}.rst".format(name), "w") as f:
        f.write(text)

        if examples:
            if len(examples) > 1:
                f.write("Examples:\n---------")
            else:
                f.write("Example:\n--------")

            for root_lexicon, text in examples:
                tree = parce.root(root_lexicon, text)
                buf = io.StringIO()
                tree.dump(buf)
                f.write(EXAMPLE_STUB.format(root_lexicon, name, indent(text),
                                            indent(buf.getvalue())))


def get_examples():
    """Return a list of all example documents in tests/lang/example.*."""
    pattern = os.path.join(EXAMPLES_DIRECTORY, "example*.*")
    return glob.glob(pattern)


def main():

    # make a mapping from the example files to their Language
    all_examples = collections.defaultdict(list)
    for filename in get_examples():
        text = open(filename).read()    # TODO encoding?
        root_lexicon = parce.find(filename=filename, contents=text)
        all_examples[root_lexicon.language].append((root_lexicon, text))


    with open("source/langs.inc", "w") as f:
        f.write(LANGS_INC_HEADER)
        for name in get_all_modules():
            langs = list(get_languages(name))
            if langs:
                clss = ", ".join(":class:`~parce.lang.{0}.{1}`".format(name, lang.__name__)
                    for lang in langs)
                examples = sum((all_examples.get(lang, []) for lang in langs), [])
                make_stub(name, examples)
                f.write("   * - :mod:`~parce.lang.{0}`\n".format(name))
                f.write("     - {0}\n\n".format(clss))

if __name__ == "__main__":
    main()




# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
A Document mixin that keeps all text tokenized.

When the text is modified, retokenizes only the modified part.
"""


import parce.treebuilder


class TreeDocumentMixin:
    """Encapsulates a full tokenized text string.

    Combine this class with a subclass of AbstractDocument (see the
    :py:mod:`document <parce.document>` module).

    Everytime the text is modified, only the modified part is retokenized. If
    that changes the lexicon in which the last part (after the modified part)
    starts, that part is also retokenized, until the state (the list of active
    lexicons) matches the state of existing tokens.

    """

    def __init__(self, builder):
        """Initialize with a TreeBuilder instance, which is doing the work."""
        self._builder = builder

    def builder(self):
        """Return the TreeBuilder we were instantiated with."""
        return self._builder

    def get_root(self, wait=False, callback=None, args=(), kwargs={}):
        """Get the root element of the completed tree.

        If wait is True, this call blocks until tokenizing is done, and the
        full tree is returned. If wait is False, None is returned if the tree
        is still busy being built.

        If a callback is given and a BackgroundTreeBuilder was used and
        tokenizing is still busy, that callback is called once when tokenizing
        is ready. If given, args and kwargs are the arguments the callback is
        called with, defaulting to () and {}, respectively.

        Note that, for the lifetime of a Document, the root element is always
        the same. But using this method you can be sure that you are dealing
        with a complete and fully intact tree.

        """
        return self.builder().get_root(wait, callback, args, kwargs)

    def root_lexicon(self):
        """Return the currently set root lexicon."""
        return self.builder().root.lexicon

    def set_root_lexicon(self, root_lexicon):
        """Set the root lexicon to use to tokenize the text."""
        self.builder().rebuild(self.text(), root_lexicon)

    def open_lexicons(self):
        """Return the list of lexicons that were left open at the end of the text.

        The root lexicon is not included; if parsing ended in the root lexicon,
        this list is empty, and the text can be considered "complete."

        """
        return self.builder().lexicons

    def modified_range(self):
        """Return a two-tuple(start, end) describing the range that was re-tokenized."""
        b = self.builder()
        return b.start, b.end

    def contents_changed(self, start, removed, added):
        """Called after modification of the text, retokenizes the modified part."""
        self.builder().rebuild(self.text(), False, start, removed, added)
        super().contents_changed(start, removed, added)




# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Test proper functioning of all examples in the ./lang directory.
"""

EXAMPLES_DIRECTORY = "tests/lang/"

import glob
import os
import sys

sys.path.insert(0, ".")

import parce


def get_examples():
    """Return a list of all example documents in tests/lang/example.*."""
    pattern = os.path.join(EXAMPLES_DIRECTORY, "example*.*")
    return glob.glob(pattern)


def test_main():
    for filename in get_examples():
        print("Testing:", filename)
        text = open(filename).read()    # TODO encoding?
        root_lexicon = parce.find(filename=filename, contents=text)
        assert root_lexicon
        print("Root lexicon: {}, result tree:".format(root_lexicon))
        parce.root(root_lexicon, text).dump()


if __name__ == "__main__":
    test_main()
